import base64
import collections
import datetime
import functools
import re
import time

import boto3
import yaml
from botocore.exceptions import ClientError
from click import FileError

from .manaus.boto_proxy import BotoClientProxy
from .manaus.utils import extract_client_error_code
from .stack_references import check_file_exceptions


def resolve_referenced_resource(ref: dict, region: str):
    if 'Stack' in ref and 'LogicalId' in ref:
        cf = BotoClientProxy('cloudformation', region)
        resource = cf.describe_stack_resource(
            StackName=ref['Stack'],
            LogicalResourceId=ref['LogicalId'])['StackResourceDetail']
        if not is_status_complete(resource['ResourceStatus']):
            raise ValueError('Resource "{}" is not ready ("{}")'.format(ref['LogicalId'], resource['ResourceStatus']))

        resource_id = resource['PhysicalResourceId']

        # sg is referenced by its name not its id
        if resource['ResourceType'] == 'AWS::EC2::SecurityGroup':
            sg = get_security_group(region, resource_id)
            return sg.id if sg is not None else None
        else:
            return resource_id
    elif 'Stack' in ref and 'Output' in ref:
        cf = BotoClientProxy('cloudformation', region)
        stack = cf.describe_stacks(
            StackName=ref['Stack'])['Stacks'][0]
        if not is_status_complete(stack['StackStatus']):
            raise ValueError('Stack "{}" is not ready ("{}")'.format(ref['Stack'], stack['StackStatus']))

        for output in stack.get('Outputs', []):
            if output['OutputKey'] == ref['Output']:
                return output['OutputValue']

        return None
    else:
        return ref


def is_status_complete(status: str):
    return status in ('CREATE_COMPLETE', 'UPDATE_COMPLETE')


def get_security_group(region: str, sg_name: str):
    ec2 = boto3.resource('ec2', region)
    try:
        # first try by tag name then by group-name (cannot be changed)
        for _filter in [{'Name': 'tag:Name', 'Values': [sg_name]}, {'Name': 'group-name', 'Values': [sg_name]}]:
            sec_groups = list(ec2.security_groups.filter(Filters=[_filter]))
            if sec_groups:
                # FIXME: What if we have 2 VPC, with a SG with the same name?!
                return sec_groups[0]
    except ClientError as e:
        error_code = extract_client_error_code(e)
        if error_code == 'InvalidGroup.NotFound':
            return None
        elif error_code == 'VPCIdNotSpecified':
            # no Default VPC, we must use the lng way...
            for sg in ec2.security_groups.all():
                # FIXME: What if we have 2 VPC, with a SG with the same name?!
                if sg.group_name == sg_name:
                    return sg
            return None
        else:
            raise


def get_vpc_attribute(region: str, vpc_id: str, attribute: str):
    ec2 = boto3.resource('ec2', region)
    vpc = ec2.Vpc(vpc_id)

    return getattr(vpc, attribute, None)


def encrypt(region: str, KeyId: str, Plaintext: str, b64encode=False):
    kms = BotoClientProxy('kms', region)
    encrypted = kms.encrypt(KeyId=KeyId, Plaintext=Plaintext)['CiphertextBlob']
    if b64encode:
        return base64.b64encode(encrypted).decode('utf-8')

    return encrypted


def list_kms_keys(region: str, details=True):
    kms = BotoClientProxy('kms', region)
    keys = list(kms.list_keys()['Keys'])
    if details:
        aliases = kms.list_aliases()['Aliases']

        for key in keys:
            key['aliases'] = [a['AliasName'] for a in aliases if a.get('TargetKeyId') == key['KeyId']]
            key.update(kms.describe_key(KeyId=key['KeyId'])['KeyMetadata'])

    return keys


def resolve_security_group(security_group, region: str):
    if isinstance(security_group, dict):
        sg = resolve_referenced_resource(security_group, region)
        if not sg:
            raise ValueError('Referenced Security Group "{}" does not exist'.format(security_group))
        return sg
    elif security_group.startswith('sg-'):
        return security_group
    else:
        sg = get_security_group(region, security_group)
        if not sg:
            raise ValueError('Security Group "{}" does not exist'.format(security_group))
        return sg.id


def resolve_security_groups(security_groups: list, region: str):
    result = []
    for security_group in security_groups:
        result.append(resolve_security_group(security_group, region))
    return result


def parse_time(s: str) -> float:
    """
    Parses an ISO 8601 string and returns a timestamp
    """
    try:
        utc = datetime.datetime.strptime(s, '%Y-%m-%dT%H:%M:%S.%fZ')
        ts = time.time()
        utc_offset = datetime.datetime.fromtimestamp(ts) - datetime.datetime.utcfromtimestamp(ts)
        local = utc + utc_offset
        return local.timestamp()
    except:
        return None


def get_required_capabilities(data: dict):
    """
    Get capabilities for a given cloud formation template for the
    "create_stack" call
    """
    capabilities = []
    for logical_id, config in data.get('Resources', {}).items():
        if config.get('Type').startswith('AWS::IAM'):
            capabilities.append('CAPABILITY_IAM')
    return capabilities


def resolve_topic_arn(region, topic_name):
    topic_arn = None
    if topic_name.startswith('arn:'):
        topic_arn = topic_name
    else:
        # resolve topic name to ARN
        sns = boto3.resource('sns', region)
        for topic in sns.topics.all():
            if topic.arn.endswith(':{}'.format(topic_name)):
                topic_arn = topic.arn

    return topic_arn


@functools.total_ordering
class SenzaStackSummary:
    def __init__(self, stack):
        self.stack = stack
        parts = stack['StackName'].rsplit('-', 1)
        self.name = parts[0]
        if len(parts) > 1:
            self.version = parts[1]
        else:
            self.version = ''

    def __getattr__(self, item):
        if item in self.__dict__:
            return self.__dict__[item]
        return self.stack.get(item)

    def __lt__(self, other):
        def key(v):
            return (v.name, v.version)
        return key(self) < key(other)

    def __eq__(self, other):
        return self.stack['StackName'] == other.stack['StackName']


def get_stacks(stack_refs: list, region, all=False, unique_only=False):
    # boto3.resource('cf')-stacks.filter() doesn't support status_filter, only StackName
    cf = BotoClientProxy('cloudformation', region)
    if all:
        status_filter = []
    else:
        # status_filter = [st for st in cf.valid_states if st != 'DELETE_COMPLETE']
        status_filter = [
            "CREATE_IN_PROGRESS",
            "CREATE_FAILED",
            "CREATE_COMPLETE",
            "ROLLBACK_IN_PROGRESS",
            "ROLLBACK_FAILED",
            "ROLLBACK_COMPLETE",
            "DELETE_IN_PROGRESS",
            "DELETE_FAILED",
            # "DELETE_COMPLETE",
            "UPDATE_IN_PROGRESS",
            "UPDATE_COMPLETE_CLEANUP_IN_PROGRESS",
            "UPDATE_COMPLETE",
            "UPDATE_ROLLBACK_IN_PROGRESS",
            "UPDATE_ROLLBACK_FAILED",
            "UPDATE_ROLLBACK_COMPLETE_CLEANUP_IN_PROGRESS",
            "UPDATE_ROLLBACK_COMPLETE"
        ]
    kwargs = {'StackStatusFilter': status_filter}
    stacks = []
    while 'NextToken' not in kwargs or kwargs['NextToken']:
        results = cf.list_stacks(**kwargs)
        for stack in results['StackSummaries']:
            if not stack_refs or matches_any(stack['StackName'], stack_refs):
                stacks.append(stack)
        kwargs['NextToken'] = results.get('NextToken')
    # After going through all stacks
    check_file_exceptions(stack_refs)

    stacks.sort(key=lambda x: x['CreationTime'], reverse=True)
    # stack names that were already yielded to avoid yielding old deleted
    # stacks whose name was reused
    stacks_yielded = set()
    for stack in stacks:
        if stack['StackName'] not in stacks_yielded or not unique_only:
            stacks_yielded.add(stack['StackName'])
            yield SenzaStackSummary(stack)


def matches_any(cf_stack_name: str, stack_refs: list):
    """
    Checks if the stack name matches any of the stack references
    """

    cf_stack_name = cf_stack_name or ''  # ensure cf_stack_name is a str
    try:
        name, version = cf_stack_name.rsplit('-', 1)
    except ValueError:
        name = cf_stack_name
        version = ""
    return any(ref.matches(name, version) for ref in stack_refs)


def get_tag(tags: list, key: str, default=None):
    """
    Get value for tag from the [{"Key": key, "Value": value}] format returned
    by AWS
    """
    if isinstance(tags, list):
        found = [tag['Value'] for tag in tags if tag['Key'] == key]
        if len(found):
            return found[0]
    return default


def get_account_id():
    conn = BotoClientProxy('iam')
    try:
        own_user = conn.get_user()['User']
    except:
        own_user = None
    if not own_user:
        roles = conn.list_roles()['Roles']
        if not roles:
            users = conn.list_users()['Users']
            if not users:
                saml = conn.list_saml_providers()['SAMLProviderList']
                if not saml:
                    return None
                else:
                    arn = [s['Arn'] for s in saml][0]
            else:
                arn = [u['Arn'] for u in users][0]
        else:
            arn = [r['Arn'] for r in roles][0]
    else:
        arn = own_user['Arn']
    account_id = arn.split(':')[4]
    return account_id


def get_account_alias():
    conn = BotoClientProxy('iam')
    return conn.list_account_aliases()['AccountAliases'][0]


class StackReference(collections.namedtuple('StackReference', 'name version')):
    def __init__(self, *args, **kwargs):
        self.matched = 0
        self.possible_definition_file = (self.name.endswith('.yml') or
                                         self.name.endswith('.yaml'))

    def raise_file_exception(self):
        """
        If it looks like a filename and didn't match anything try to open it
        to see if exists and can be opened and raise an exception if it can't
        """
        if not self.matched and self.possible_definition_file:
            try:
                with open(self.name) as fd:
                    data = yaml.safe_load(fd)
                assert data['SenzaInfo']['StackName']
            except (OSError, IOError) as error:
                raise FileError(self.name, error.strerror)
            except KeyError as error:
                raise ValueError("SenzaInfo.StackName missing from definition file")

    def matches(self, name: str, version: str):
        matches_name = re.match(self.name + '$', name)
        matches_version = (not self.version or
                           re.match(self.version + '$', version))
        matches = bool(matches_name and matches_version)
        if matches:
            self.matched += 1
        return matches

    def cf_stack_name(self):
        return ('{}-{}'.format(self.name, self.version)
                if self.version
                else self.name)



#!/usr/bin/env python3
import base64
import calendar
import collections
import configparser
import datetime
import ipaddress
import json
import os
import re
import sys
import time
from pprint import pformat
from urllib.error import URLError
from urllib.parse import quote
from urllib.request import urlopen

import boto3
import click
import requests
import yaml
from botocore.exceptions import ClientError
from clickclick import (Action, FloatRange, OutputFormat, choice, error,
                        fatal_error, info, ok)
from clickclick.console import print_table

from .arguments import (GLOBAL_OPTIONS, json_output_option, output_option,
                        parameter_file_option, region_option,
                        stacktrace_visible_option, watch_option,
                        watchrefresh_option)
from .aws import (StackReference, get_account_alias, get_account_id,
                  get_required_capabilities, get_stacks, get_tag, matches_any,
                  parse_time, resolve_topic_arn)
from .components import evaluate_template, get_component
from .components.stups_auto_configuration import find_taupage_image
from .error_handling import HandleExceptions
from .exceptions import InvalidDefinition
from .manaus.boto_proxy import BotoClientProxy
from .manaus.cloudformation import CloudFormation
from .manaus.ec2 import EC2
from .manaus.exceptions import VPCError
from .manaus.route53 import Route53, Route53Record
from .manaus.utils import extract_client_error_code
from .patch import patch_auto_scaling_group
from .respawn import get_auto_scaling_group, respawn_auto_scaling_group
from .stups.piu import Piu
from .subcommands.config import cmd_config
from .subcommands.root import cli
from .templates import get_template_description, get_templates
from .templates._helper import get_mint_bucket_name
from .traffic import (change_version_traffic, get_records,
                      print_version_traffic, resolve_to_ip_addresses)
from .utils import (camel_case_to_underscore, ensure_keys, named_value,
                    pystache_render)

STYLES = {
    'RUNNING': {'fg': 'green'},
    'TERMINATED': {'fg': 'red'},
    'DELETE_COMPLETE': {'fg': 'red'},
    'DELETE_FAILED': {'fg': 'red'},
    'ROLLBACK_COMPLETE': {'fg': 'red'},
    'CREATE_COMPLETE': {'fg': 'green'},
    'CREATE_FAILED': {'fg': 'red'},
    'PENDING': {'fg': 'yellow', 'bold': True},
    'CREATE_IN_PROGRESS': {'fg': 'yellow', 'bold': True},
    'DELETE_IN_PROGRESS': {'fg': 'red', 'bold': True},
    'STOPPING': {'fg': 'red'},
    'STOPPED': {'fg': 'red', 'bold': True},
    'SHUTTING_DOWN': {'fg': 'red', 'bold': True},
    'ROLLBACK_IN_PROGRESS': {'fg': 'red', 'bold': True},
    'ROLLBACK_FAILED': {'fg': 'red'},
    'UPDATE_COMPLETE': {'fg': 'green'},
    'UPDATE_ROLLBACK_IN_PROGRESS': {'fg': 'red', 'bold': True},
    'UPDATE_IN_PROGRESS': {'fg': 'yellow', 'bold': True},
    'UPDATE_ROLLBACK_COMPLETE_CLEANUP_IN_PROGRESS': {'fg': 'red', 'bold': True},
    'UPDATE_COMPLETE_CLEANUP_IN_PROGRESS': {'fg': 'yellow', 'bold': True},
    'UPDATE_FAILED': {'fg': 'red'},
    'UPDATE_ROLLBACK_COMPLETE': {'fg': 'red'},
    'IN_SERVICE': {'fg': 'green'},
    'HEALTHY': {'fg': 'green'},
    'OUT_OF_SERVICE': {'fg': 'red'},
    'OK': {'fg': 'green'},
    'ERROR': {'fg': 'red'},
}


TITLES = {
    'creation_time': 'Created',
    'LogicalResourceId': 'Resource ID',
    'launch_time': 'Launched',
    'ResourceStatus': 'Status',
    'ResourceStatusReason': 'Status Reason',
    'lb_status': 'LB Status',
    'private_ip': 'Private IP',
    'public_ip': 'Public IP',
    'resource_id': 'Resource ID',
    'instance_id': 'Instance ID',
    'version': 'Ver.',
    'total_instances': 'Inst.#',
    'running_instances': 'Running',
    'docker_source': 'Docker Image Source',
    'healthy_instances': 'Healthy',
    'http_status': 'HTTP',
    'main_dns': 'Main DNS',
    'id': 'ID',
    'ImageId': 'Image ID',
    'OwnerId': 'Owner',
    'healthy_hosts': 'Healthy #',
    'requests_per_sec': 'Req./s',
    '4xx_percentage': '4xx %',
    '5xx_percentage': '5xx %',
    'latency_ms': 'Latency (ms)',
}

MAX_COLUMN_WIDTHS = {
    'description': 50,
    'stacks': 20,
    'ResourceStatusReason': 50
}


def print_json(data, output=None):
    if output == 'yaml':
        parsed_data = yaml.safe_load(data)
        print(yaml.safe_dump(parsed_data, indent=4, default_flow_style=False))
    else:
        print(data)


class DefinitionParamType(click.ParamType):
    name = 'definition'

    def convert(self, value, param, ctx):
        if isinstance(value, str):
            try:
                url = value if '://' in value else 'file://{}'.format(quote(os.path.abspath(value)))
                # if '://' not in value:
                #     url = 'file://{}'.format(quote(os.path.abspath(value)))

                response = urlopen(url)
                try:
                    data = yaml.safe_load(response.read())
                except yaml.YAMLError as e:
                    raise InvalidDefinition(path=value,
                                            reason=str(e))
            except URLError:
                self.fail('"{}" not found'.format(value), param, ctx)
        else:
            data = value

        if 'SenzaInfo' not in data:
            self.fail('"SenzaInfo" entry is missing in '
                      'YAML file "{}"'.format(value),
                      param, ctx)
        return data


class KeyValParamType(click.ParamType):
    name = 'key_val'

    def convert(self, value, param, ctx):
        if isinstance(value, str):
            try:
                key, val = value.split('=', 1)
            except:
                self.fail('invalid key value parameter "{}" (must be KEY=VAL)'.format(value))
            key_val = (key, val)
        else:
            key_val = value
        return key_val


def validate_version(ctx, param, value):
    if not VERSION_PATTERN.match(value):
        raise click.BadParameter('Version must satisfy regular expression pattern "[a-zA-Z0-9]+"')
    return value


def watching(w: bool, watch: int):
    if w and not watch:
        watch = 2
    if watch:
        click.clear()
    yield 0
    if watch:
        while True:
            time.sleep(watch)
            click.clear()
            yield 0


# from AWS docs:
# Stack name must contain only alphanumeric characters (case sensitive)
# and start with an alpha character. Maximum length of the name is 255 characters.
STACK_NAME_PATTERN = re.compile(r'^[a-zA-Z][a-zA-Z0-9-]*$')
VERSION_PATTERN = re.compile(r'^[a-zA-Z0-9]+$')

DEFINITION = DefinitionParamType()

KEY_VAL = KeyValParamType()

BASE_TEMPLATE = {
    'AWSTemplateFormatVersion': '2010-09-09'
}


def evaluate(definition, args, account_info, force: bool):
    # extract Senza* meta information
    info = definition.pop("SenzaInfo")
    info["StackVersion"] = args.version
    # replace Arguments and AccountInfo Variabales in info section
    info = yaml.load(evaluate_template(yaml.dump(info), {}, {}, args, account_info))

    # add info as mappings
    # http://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/mappings-section-structure.html
    definition = ensure_keys(definition, "Mappings", "Senza", "Info")
    definition["Mappings"]["Senza"]["Info"] = info

    template = yaml.dump(definition, default_flow_style=False)
    definition = evaluate_template(template, info, [], args, account_info)
    definition = yaml.load(definition)

    components = definition.pop("SenzaComponents", [])

    # merge base template with definition
    BASE_TEMPLATE.update(definition)
    definition = BASE_TEMPLATE

    # evaluate all components
    for component in components:
        componentname, configuration = named_value(component)
        configuration["Name"] = componentname

        componenttype = configuration["Type"]
        componentfn = get_component(componenttype)

        if not componentfn:
            raise click.UsageError('Component "{}" does not exist'.format(componenttype))

        definition = componentfn(definition, configuration, args, info, force, account_info)

    # throw executed template to templating engine and provide all information for substitutions
    template = yaml.dump(definition, default_flow_style=False)
    definition = evaluate_template(template, info, components, args, account_info)
    definition = yaml.load(definition)

    return definition


class TemplateArguments:
    def __init__(self, **kwargs):
        for key, val in kwargs.items():
            setattr(self, key, val)


class AccountArguments:
    """
    Account arguments to use in the definitions
    """
    def __init__(self, region):
        self.Region = region
        self.__AccountAlias = None
        self.__AccountID = None
        self.__Domain = None
        self.__MintBucket = None
        self.__TeamID = None
        self.__VpcID = None

    @property
    def AccountID(self):
        if self.__AccountID is None:
            self.__AccountID = get_account_id()
        return self.__AccountID

    @property
    def AccountAlias(self):
        if self.__AccountAlias is None:
            self.__AccountAlias = get_account_alias()
        return self.__AccountAlias

    @property
    def Domain(self) -> str:
        if self.__Domain is None:
            self.__setDomain()
        return self.__Domain.rstrip('.')

    def __setDomain(self, domain_name=None):
        domain_list = list(Route53.get_hosted_zones(domain_name))
        if len(domain_list) == 0:
            raise AttributeError('No Domain configured')
        elif len(domain_list) > 1:
            domain = choice('Please select the domain',
                            sorted(domain.domain_name
                                   for domain in domain_list))
        else:
            domain = domain_list[0].domain_name
        self.__Domain = domain
        return domain

    def split_domain(self, domain_name):
        self.__setDomain(domain_name)
        if domain_name.endswith('.{}'.format(self.Domain)):
            return domain_name[:-len('.{}'.format(self.Domain))], self.Domain
        else:
            # default behaviour for unknown domains
            return domain_name.split('.', 1)

    @property
    def TeamID(self):
        if self.__TeamID is None:
            self.__TeamID = self.AccountAlias.split('-', maxsplit=1)[-1]
        return self.__TeamID

    @property
    def VpcID(self):
        if self.__VpcID is None:
            ec2 = EC2(self.Region)
            try:
                vpc = ec2.get_default_vpc()
            except VPCError as error:
                if sys.stdin.isatty() and error.number_of_vpcs:
                    # if running in interactive terminal and there are VPCs
                    # to choose from
                    vpcs = ec2.get_all_vpcs()
                    options = [(vpc.vpc_id, str(vpc)) for vpc in vpcs]
                    print("Can't find a default VPC")
                    vpc = choice("Select VPC to use",
                                 options=options)
                else:  # if not running in interactive terminal (e.g Jenkins)
                    raise
            self.__VpcID = vpc.vpc_id
        return self.__VpcID

    @property
    def MintBucket(self):
        if self.__MintBucket is None:
            self.__MintBucket = get_mint_bucket_name(self.Region)
        return self.__MintBucket


def parse_args(input, region, version, parameter, account_info):
    paras = {}
    defaults = collections.OrderedDict()
    parameterlist = []

    # process positional parameters first
    seen_keyword = False
    for i, param in enumerate(input['SenzaInfo'].get('Parameters', [])):
        for key, config in param.items():
            parameterlist.append(key)
            # collect all allowed keys and default values regardless
            paras[key] = None
            defaults[key] = config.get('Default', None)
            if defaults[key] is not None:
                defaults[key] = pystache_render(str(defaults[key]), {'AccountInfo': account_info})
            if i < len(parameter):
                if '=' in parameter[i]:
                    seen_keyword = True
                else:
                    if seen_keyword:
                        raise click.UsageError("Positional parameters must not follow keywords.")
                    paras[key] = parameter[i]

    if len(paras) < len(parameter):
        raise click.UsageError('Too many parameters given. Need only: "{}"'.format(' '.join(parameterlist)))

    # process keyword parameters separately, if any
    if seen_keyword:
        for param in parameter:
            if '=' in param:
                key, value = param.split('=', 1)  # split only on first =
                if key not in paras:
                    raise click.UsageError('Unrecognized keyword parameter: "{}"'.format(key))
                if paras[key] is not None:
                    raise click.UsageError('Parameter specified multiple times: "{}"'.format(key))
                paras[key] = value

    # finally, make sure every parameter got a value assigned, using defaults if given
    for key, defval in defaults.items():
        paras[key] = paras[key] or defval
        if paras[key] is None:
            raise click.UsageError('Missing parameter "{}". Need: "{}"'.format(key, ' '.join(parameterlist)))

    args = TemplateArguments(region=region, version=version, **paras)
    return args


def read_parameter_file(parameter_file):
    paras = []

    try:
        url = (parameter_file if '://' in parameter_file
               else 'file://{}'.format(quote(os.path.abspath(parameter_file))))

        response = urlopen(url)
    except URLError:
        raise click.UsageError('Can\'t read parameter file "{}"'.format(parameter_file))

    try:
        cfg = yaml.safe_load(response.read())
        for key, val in cfg.items():
            paras.append("{}={}".format(key, val))
    except yaml.YAMLError as e:
        raise click.UsageError('Error {}'.format(e))

    return tuple(paras)


def get_region(region):
    if not region:
        region = GLOBAL_OPTIONS.get('region')
    if not region:
        config = configparser.ConfigParser()
        try:
            config.read(os.path.expanduser('~/.aws/config'))
            if 'default' in config:
                region = config['default']['region']
        except:
            pass

    if not region:
        raise click.UsageError('Please specify the AWS region on the '
                               'command line (--region) or in ~/.aws/config')

    return region


def check_credentials(region):
    iam = BotoClientProxy('iam')
    return iam.list_account_aliases()


def get_stack_refs(refs: list):
    """
    Converts a list of strings in a list of StackReferences
    """

    refs = list(refs)
    refs.reverse()
    stack_refs = []
    last_stack = None
    while refs:
        ref = refs.pop()

        if last_stack is not None and re.compile(r'v[0-9][a-zA-Z0-9-]*$').match(ref):
            stack_refs.append(StackReference(last_stack, ref))
        else:
            try:
                with open(ref) as fd:
                    try:
                        data = yaml.safe_load(fd)
                    except yaml.YAMLError as e:
                        raise InvalidDefinition(path=ref,
                                                reason=str(e))

                try:
                    ref = data['SenzaInfo']['StackName']
                except KeyError:
                    raise InvalidDefinition(path=ref,
                                            reason="SenzaInfo is missing "
                                                   "or invalid")
                except TypeError:
                    raise InvalidDefinition(path=ref,
                                            reason="Invalid SenzaInfo")
            except (OSError, IOError):
                # It's still possible that the ref is a regex
                pass

            if refs:
                version = refs.pop()
            else:
                version = None
            stack_refs.append(StackReference(ref, version))
            last_stack = ref
    return stack_refs


def all_with_version(stack_refs: list):
    """
    Checks if all stackreferences have versions
    """
    for ref in stack_refs:
        if not ref.version:
            return False
    return True


@cli.command('list')
@region_option
@output_option
@watch_option
@watchrefresh_option
@click.option('--all', is_flag=True, help='Show all stacks, including deleted ones')
@click.argument('stack_ref', nargs=-1)
@stacktrace_visible_option
def list_stacks(region, stack_ref, all, output, w, watch):
    '''List Cloud Formation stacks'''

    region = get_region(region)
    check_credentials(region)

    stack_refs = get_stack_refs(stack_ref)

    for _ in watching(w, watch):
        rows = []
        for stack in get_stacks(stack_refs, region, all=all):
            rows.append({'stack_name': stack.name,
                         'version': stack.version,
                         'status': stack.StackStatus,
                         'creation_time': calendar.timegm(stack.CreationTime.timetuple()),
                         'description': stack.TemplateDescription})

        rows.sort(key=lambda x: (x['stack_name'], x['version']))

        with OutputFormat(output):
            print_table('stack_name version status creation_time description'.split(), rows,
                        styles=STYLES, titles=TITLES)


def get_application_load_balancer_metrics(cloudwatch, alb_id, start, now):
    alb_metrics = {
        'TargetResponseTime': 'Average', 'RequestCount': 'Sum',
        'HTTPCode_Target_5XX_Count': 'Sum', 'HTTPCode_Target_4XX_Count': 'Sum'}
    row = {}
    data = {}
    for k, v in alb_metrics.items():
        res = cloudwatch.get_metric_statistics(
            Namespace='AWS/ApplicationELB',
            MetricName=k,
            Dimensions=[{'Name': 'LoadBalancer', 'Value': alb_id}],
            StartTime=start,
            EndTime=now,
            Period=60,
            Statistics=[v])
        # NOTE: we get the most recent two (!) datapoints as the newest point might be incomplete
        most_recent = sorted(res['Datapoints'], key=lambda x: x['Timestamp'])[-2:]
        if most_recent:
            data[(k, v)] = most_recent[0][v]
        else:
            data[(k, v)] = None
    if data[('RequestCount', 'Sum')] is not None:
        requests_per_min = data[('RequestCount', 'Sum')]
        row['requests_per_sec'] = round(requests_per_min / 60, 2)
    else:
        requests_per_min = 0
    if data[('HTTPCode_Target_4XX_Count', 'Sum')] is not None:
        row['4xx_percentage'] = round((data[('HTTPCode_Target_4XX_Count', 'Sum')]) /
                                      (requests_per_min + 0.0001), 2)
    if data[('HTTPCode_Target_5XX_Count', 'Sum')] is not None:
        row['5xx_percentage'] = round((data[('HTTPCode_Target_5XX_Count', 'Sum')]) /
                                      (requests_per_min + 0.0001), 2)
    if data[('TargetResponseTime', 'Average')] is not None:
        row['latency_ms'] = int(round((data[('TargetResponseTime', 'Average')]) * 1000, 0))
    return row


def get_classic_load_balancer_metrics(cloudwatch, lb_name, start, now):
    elb_metrics = {
        'Latency': 'Average', 'RequestCount': 'Sum',
        'HTTPCode_Backend_5XX': 'Sum', 'HTTPCode_Backend_4XX': 'Sum'}
    row = {}
    data = {}
    for k, v in elb_metrics.items():
        res = cloudwatch.get_metric_statistics(
            Namespace='AWS/ELB',
            MetricName=k,
            Dimensions=[{'Name': 'LoadBalancerName', 'Value': lb_name}],
            StartTime=start,
            EndTime=now,
            Period=60,
            Statistics=[v])
        most_recent = sorted(res['Datapoints'], key=lambda x: x['Timestamp'])[-1:]
        if most_recent:
            data[(k, v)] = most_recent[0][v]
        else:
            data[(k, v)] = None
    if data[('RequestCount', 'Sum')] is not None:
        requests_per_min = data[('RequestCount', 'Sum')]
        row['requests_per_sec'] = round(requests_per_min / 60, 2)
    else:
        requests_per_min = 0
    if data[('HTTPCode_Backend_4XX', 'Sum')] is not None:
        row['4xx_percentage'] = round((data[('HTTPCode_Backend_4XX', 'Sum')]) /
                                      (requests_per_min + 0.0001), 2)
    if data[('HTTPCode_Backend_5XX', 'Sum')] is not None:
        row['5xx_percentage'] = round((data[('HTTPCode_Backend_5XX', 'Sum')]) /
                                      (requests_per_min + 0.0001), 2)
    if data[('Latency', 'Average')] is not None:
        row['latency_ms'] = int(round((data[('Latency', 'Average')]) * 1000, 0))
    return row


@cli.command()
@region_option
@output_option
@watch_option
@watchrefresh_option
@click.option('--all', is_flag=True, help='Show all stacks, including deleted ones')
@click.argument('stack_ref', nargs=-1)
@stacktrace_visible_option
def health(region, stack_ref, all, output, w, watch):
    '''Show stack health (ELB req/s, ..)'''

    region = get_region(region)
    check_credentials(region)

    cloudwatch = BotoClientProxy('cloudwatch', region)

    stack_refs = get_stack_refs(stack_ref)

    for _ in watching(w, watch):
        rows = []
        start = datetime.datetime.utcnow() - datetime.timedelta(minutes=5)
        now = datetime.datetime.utcnow()
        paginator = cloudwatch.get_paginator('list_metrics')
        alb_ids = {}
        for page in paginator.paginate(Namespace='AWS/ApplicationELB', MetricName='RequestCount',
                                       Dimensions=[{'Name': 'LoadBalancer'}]):
            for metric in page['Metrics']:
                alb_id = ''.join([d['Value'] for d in metric['Dimensions'] if d['Name'] == 'LoadBalancer'])
                alb_ids[alb_id.split('/')[1]] = alb_id
        for stack in get_stacks(stack_refs, region, all=all):
            lb_name = stack.StackName
            instance_health = get_instance_health(lb_name, region)
            row = {
                'stack_name': stack.name,
                'version': stack.version,
                'status': stack.StackStatus,
                'healthy_hosts': get_healthy_instances(instance_health),
                'creation_time': calendar.timegm(stack.CreationTime.timetuple()),
            }
            alb_id = alb_ids.get(lb_name)
            if alb_id:
                row.update(get_application_load_balancer_metrics(cloudwatch, alb_id, start, now))
            else:
                row.update(get_classic_load_balancer_metrics(cloudwatch, lb_name, start, now))
            rows.append(row)

        rows.sort(key=lambda x: (x['stack_name'], x['version']))

        with OutputFormat(output):
            print_table('stack_name version status healthy_hosts requests_per_sec '
                        '4xx_percentage 5xx_percentage latency_ms'.split(), rows,
                        styles=STYLES, titles=TITLES)


@cli.command()
@click.argument('definition', type=DEFINITION)
@click.argument('version', callback=validate_version)
@click.argument('parameter', nargs=-1)
@region_option
@parameter_file_option
@click.option('--disable-rollback', is_flag=True, help='Disable Cloud Formation rollback on failure')
@click.option('--dry-run', is_flag=True, help='No-op mode: show what would be created')
@click.option('-f', '--force', is_flag=True, help='Ignore failing validation checks')
@click.option('-t', '--tag', help='Tags to associate with the stack.', multiple=True)
@stacktrace_visible_option
def create(definition, region, version, parameter, disable_rollback, dry_run,
           force, tag, parameter_file):
    '''Create a new Cloud Formation stack from the given Senza definition file'''

    region = get_region(region)
    data = create_cf_template(definition, region, version, parameter, force, parameter_file)

    for tag_kv in tag:
        try:
            key, value = tag_kv.split('=')
        except ValueError:
            fatal_error('Invalid tag {}. Tags should be in the form of key=value'.format(tag_kv))
        data['Tags'].append({'Key': key, 'Value': value})

    cf = BotoClientProxy('cloudformation', region)

    with Action('Creating Cloud Formation stack {}..'.format(data['StackName'])) as act:
        try:
            if dry_run:
                info('**DRY-RUN** {}'.format(data['NotificationARNs']))
                info(' Tags: {}'.format(data['Tags']))
            else:
                cf.create_stack(DisableRollback=disable_rollback, **data)
        except ClientError as e:
            if extract_client_error_code(e) == 'AlreadyExistsException':
                act.fatal_error('Stack {} already exists. Please choose another version.'.format(data['StackName']))
            else:
                raise


@cli.command()
@click.argument('definition', type=DEFINITION)
@click.argument('version', callback=validate_version)
@click.argument('parameter', nargs=-1)
@region_option
@parameter_file_option
@click.option('--disable-rollback', is_flag=True, help='Disable Cloud Formation rollback on failure')
@click.option('--dry-run', is_flag=True, help='No-op mode: show what would be created')
@click.option('-f', '--force', is_flag=True, help='Ignore failing validation checks')
@stacktrace_visible_option
def update(definition, region, version, parameter, disable_rollback, dry_run,
           force, parameter_file):
    '''Update an existing Cloud Formation stack from the given Senza definition file'''

    region = get_region(region)
    data = create_cf_template(definition, region, version, parameter, force, parameter_file)
    cf = BotoClientProxy('cloudformation', region)

    with Action('Updating Cloud Formation stack {}..'.format(data['StackName'])) as act:
        try:
            if dry_run:
                info('**DRY-RUN** {}'.format(data['NotificationARNs']))
            else:
                del(data['Tags'])
                cf.update_stack(**data)
        except ClientError as e:
            act.fatal_error('ClientError: {}'.format(pformat(e.response)))


@cli.command('print')
@click.argument('definition', type=DEFINITION)
@click.argument('version', callback=validate_version)
@click.argument('parameter', nargs=-1)
@region_option
@parameter_file_option
@json_output_option
@click.option('-f', '--force', is_flag=True, help='Ignore failing validation checks')
@stacktrace_visible_option
def print_cfjson(definition, region, version, parameter, output, force,
                 parameter_file):
    '''Print the generated Cloud Formation template'''

    region = get_region(region)
    data = create_cf_template(definition, region, version, parameter, force, parameter_file)
    print_json(data['TemplateBody'], output)


def create_cf_template(definition, region, version, parameter, force, parameter_file):
    region = get_region(region)
    if parameter_file:
        parameter_from_file = read_parameter_file(parameter_file)
        parameter = parameter + parameter_from_file
    check_credentials(region)
    account_info = AccountArguments(region=region)
    args = parse_args(definition, region, version, parameter, account_info)

    with Action('Generating Cloud Formation template..') as action:
        try:
            data = evaluate(definition.copy(), args, account_info, force)
        except VPCError as e:
            action.fatal_error('Fatal Error: {}'.format(e))
    stack_name = "{0}-{1}".format(data['Mappings']['Senza']['Info']['StackName'],
                                  data['Mappings']['Senza']['Info']['StackVersion'])
    if len(stack_name) > 128:
        fatal_error('Error: Stack name "{}" cannot exceed 128 characters. '.format(stack_name) +
                    'Please choose another name/version.')

    parameters = []
    for name, parameter in data.get("Parameters", {}).items():
        parameters.append({'ParameterKey': name, 'ParameterValue': getattr(args, name, None)})

    tags = {}
    senza_tags = data['Mappings']['Senza']['Info'].get('Tags')
    if isinstance(senza_tags, dict):
        tags.update(senza_tags)
    elif isinstance(senza_tags, list):
        for tag in senza_tags:
            for key, value in tag.items():
                # # As the SenzaInfo is not evaluated, we explicitly evaluate the values here
                tags[key] = evaluate_template(value, info, [], args, account_info)

    tags.update({
        "Name": stack_name,
        "StackName": data['Mappings']['Senza']['Info']['StackName'],
        "StackVersion": data['Mappings']['Senza']['Info']['StackVersion']
    })
    tags_list = []
    tags_mapping_list = []
    for k, v in tags.items():
        tags_list.append({'Key': k, 'Value': v})
        tags_mapping_list.append({k: v})
    data['Mappings']['Senza']['Info']['Tags'] = tags_mapping_list

    if "OperatorTopicId" in data['Mappings']['Senza']['Info']:
        topic = data['Mappings']['Senza']['Info']["OperatorTopicId"]
        topic_arn = resolve_topic_arn(region, topic)
        if not topic_arn:
            fatal_error('Error: SNS topic "{}" does not exist'.format(topic))
        topics = [topic_arn]
    else:
        topics = []

    capabilities = get_required_capabilities(data)
    cfjson = json.dumps(data, sort_keys=True, indent=4)
    return {'StackName': stack_name, 'TemplateBody': cfjson, 'Parameters': parameters, 'Tags': tags_list,
            'NotificationARNs': topics, 'Capabilities': capabilities}


@cli.command()
@click.argument('stack_ref', nargs=-1)
@region_option
@click.option('--dry-run', is_flag=True,
              help='No-op mode: show what would be deleted')
@click.option('-g', '--ignore-non-existent', is_flag=True,
              help='Do not show error when stack does not exist')
@click.option('-f', '--force', is_flag=True,
              help='Allow deleting multiple stacks and stacks with traffic')
@click.option('-i', '--interactive', is_flag=True,
              help='Prompt before every deletion')
@stacktrace_visible_option
def delete(stack_ref, region, dry_run, force, interactive, ignore_non_existent):
    '''Delete a single Cloud Formation stack'''

    stack_refs = get_stack_refs(stack_ref)
    region = get_region(region)
    check_credentials(region)

    if not stack_refs:
        raise click.UsageError('Please specify at least one stack')

    cf = CloudFormation(region=region)
    stacks = [stack
              for stack in cf.get_stacks()
              if matches_any(stack.name, stack_refs)]

    if (not all_with_version(stack_refs) and len(stacks) > 1 and
            not dry_run and not force):
        fatal_error('Error: {} matching stacks found. '.format(len(stacks)) +
                    'Please use the "--force" flag if you really want to '
                    'delete multiple stacks.')

    if not stacks and not ignore_non_existent:
        fatal_error('Error: Stack {} not found!'.format(stack_refs[0].name))

    for stack in stacks:

        for r in stack.resources:
            if isinstance(r, Route53Record):
                has_traffic = r.weight is not None and r.weight
                if has_traffic and not force:
                    fatal_error('Error: Stack {} has traffic!\n'
                                'Use --force if you really want '
                                'to delete it'.format(stack.name))

        if interactive and not click.confirm("Delete '{}'?".format(stack.name)):
            continue

        with Action('Deleting Cloud Formation stack {}..'.format(stack.name)):
            if not dry_run:
                stack.delete()


def format_resource_type(resource_type):
    if resource_type and resource_type.startswith('AWS::'):
        return resource_type[5:]
    return resource_type


@cli.command()
@click.argument('stack_ref', nargs=-1)
@region_option
@watch_option
@watchrefresh_option
@output_option
@stacktrace_visible_option
def resources(stack_ref, region, w, watch, output):
    '''Show all resources of a single Cloud Formation stack'''

    stack_refs = get_stack_refs(stack_ref)
    region = get_region(region)
    check_credentials(region)
    cf = BotoClientProxy('cloudformation', region)

    for _ in watching(w, watch):
        rows = []
        for stack in get_stacks(stack_refs, region):
            resources = cf.describe_stack_resources(StackName=stack.StackName)['StackResources']

            for resource in resources:
                d = resource.copy()
                d['stack_name'] = stack.name
                d['version'] = stack.version
                d['resource_type'] = format_resource_type(d['ResourceType'])
                d['creation_time'] = calendar.timegm(resource['Timestamp'].timetuple())
                rows.append(d)

        rows.sort(key=lambda x: (x['stack_name'], x['version'], x['LogicalResourceId']))

        with OutputFormat(output):
            print_table('stack_name version LogicalResourceId resource_type ResourceStatus creation_time'.split(),
                        rows, styles=STYLES, titles=TITLES)


@cli.command()
@click.argument('stack_ref', nargs=-1)
@region_option
@watch_option
@watchrefresh_option
@output_option
@stacktrace_visible_option
def events(stack_ref, region, w, watch, output):
    '''Show all Cloud Formation events for a single stack'''

    stack_refs = get_stack_refs(stack_ref)
    region = get_region(region)
    check_credentials(region)
    cf = BotoClientProxy('cloudformation', region)

    for _ in watching(w, watch):
        rows = []
        for stack in get_stacks(stack_refs, region):
            events = cf.describe_stack_events(StackName=stack.StackId)['StackEvents']

            for event in events:
                d = event.copy()
                d['stack_name'] = stack.name
                d['version'] = stack.version
                d['resource_type'] = format_resource_type(d['ResourceType'])
                d['event_time'] = calendar.timegm(event['Timestamp'].timetuple())
                rows.append(d)

        rows.sort(key=lambda x: x['event_time'])

        with OutputFormat(output):
            print_table(('stack_name version resource_type LogicalResourceId ' +
                        'ResourceStatus ResourceStatusReason event_time').split(),
                        rows, styles=STYLES, titles=TITLES, max_column_widths=MAX_COLUMN_WIDTHS)


@cli.command()
@click.argument('definition_file', type=click.File('w'))
@region_option
@click.option('-t', '--template', help='Use a custom template', metavar='TEMPLATE_ID')
@click.option('-v', '--user-variable', help='Provide user variables for the template',
              metavar='KEY=VAL', multiple=True, type=KEY_VAL)
@stacktrace_visible_option
def init(definition_file, region, template, user_variable):
    """Initialize a new Senza definition"""

    region = get_region(region)
    check_credentials(region)
    account_info = AccountArguments(region=region)

    templates = get_templates()

    module = templates.get(template, None)

    if not module:
        module = choice('Please select the project template',
                        [(module, get_template_description(name, module))
                         for name, module
                         in sorted(templates.items(), key=lambda x: x[0])],  # sort by key
                        default='webapp')

    variables = {}
    for key_val in user_variable:
        key, val = key_val
        variables[key] = val
    variables = module.gather_user_variables(variables, region, account_info)
    with Action('Generating Senza definition file {}..'.format(definition_file.name)):
        definition = module.generate_definition(variables)
        definition_file.write(definition)


def get_instance_health(stack_name: str, region: str) -> dict:
    if stack_name is None:
        return {}
    instance_health = {}
    elb = BotoClientProxy('elb', region)
    try:
        instance_states = elb.describe_instance_health(LoadBalancerName=stack_name)['InstanceStates']
        for istate in instance_states:
            instance_health[istate['InstanceId']] = camel_case_to_underscore(istate['State']).upper()
    except ClientError as e:
        # retry with ELBv2
        error_code = extract_client_error_code(e)
        if error_code == 'LoadBalancerNotFound':
            elbv2 = BotoClientProxy('elbv2', region)
            try:
                response = elbv2.describe_target_groups(Names=[stack_name])
                for tg in response['TargetGroups']:
                    response = elbv2.describe_target_health(TargetGroupArn=tg['TargetGroupArn'])
                    for target_health in response['TargetHealthDescriptions']:
                        instance_health[target_health['Target']['Id']] = \
                            camel_case_to_underscore(target_health['TargetHealth']['State']).upper()
            except ClientError as e:
                inner_error_code = extract_client_error_code(e)
                if inner_error_code not in ('TargetGroupNotFound', 'ValidationError', 'Throttling'):
                    raise
        # ignore ValidationError "LoadBalancer name cannot be longer than 32 characters"
        # ignore rate limit exceeded errors
        elif error_code not in ('ValidationError', 'Throttling'):
            raise
    return instance_health


def get_healthy_instances(instance_health: dict) -> int:
    if not instance_health:
        # if we have no instances at all -> treat as "unknown"
        return None
    return len([i for i in instance_health.values() if i in ('IN_SERVICE', 'HEALTHY')])


def get_instance_user_data(instance) -> dict:
    try:
        attrs = instance.describe_attribute(Attribute='userData')
        data_b64 = attrs['UserData']['Value']
        data_yaml = base64.b64decode(data_b64)
        data_dict = yaml.load(data_yaml)
        return data_dict
    except Exception as e:
        # there's just too many ways this can fail, catch 'em all
        sys.stderr.write('Failed to query instance user data: {}\n'.format(e))
    return {}


def get_instance_docker_image_source(instance) -> str:
    return get_instance_user_data(instance).get('source', '')


@cli.command()
@click.argument('stack_ref', nargs=-1)
@click.option('--all', is_flag=True, help='Show all instances, including instances not part of any stack')
@click.option('--terminated', is_flag=True, help='Show instances in TERMINATED state')
@click.option('-d', '--docker-image', is_flag=True, help='Show docker image source for every instance listed')
@click.option('-p', '--piu', metavar='REASON', help='execute PIU request-access command')
@click.option('-O', '--odd-host', help='Odd SSH bastion hostname', envvar='ODD_HOST', metavar='HOSTNAME')
@region_option
@output_option
@watch_option
@watchrefresh_option
@stacktrace_visible_option
def instances(stack_ref, all, terminated, docker_image, piu, odd_host, region,
              output, w, watch):
    '''List the stack's EC2 instances'''

    stack_refs = get_stack_refs(stack_ref)
    region = get_region(region)
    check_credentials(region)

    ec2 = boto3.resource('ec2', region)

    if all:
        filters = []
    else:
        # filter out instances not part of any stack
        filters = [{'Name': 'tag-key', 'Values': ['aws:cloudformation:stack-name']}]

    opt_docker_column = ' docker_source' if docker_image else ''

    for _ in watching(w, watch):
        rows = []

        for instance in ec2.instances.filter(Filters=filters):
            cf_stack_name = get_tag(instance.tags, 'aws:cloudformation:stack-name')
            stack_name = get_tag(instance.tags, 'StackName')
            stack_version = get_tag(instance.tags, 'StackVersion')
            if not stack_refs or matches_any(cf_stack_name, stack_refs):
                instance_health = get_instance_health(cf_stack_name, region)
                if instance.state['Name'].upper() != 'TERMINATED' or terminated:

                    docker_source = get_instance_docker_image_source(instance) if docker_image else ''

                    rows.append({'stack_name': stack_name or '',
                                 'version': stack_version or '',
                                 'resource_id': get_tag(instance.tags, 'aws:cloudformation:logical-id'),
                                 'instance_id': instance.id,
                                 'public_ip': instance.public_ip_address,
                                 'private_ip': instance.private_ip_address,
                                 'state': instance.state['Name'].upper().replace('-', '_'),
                                 'lb_status': instance_health.get(instance.id),
                                 'docker_source': docker_source,
                                 'launch_time': instance.launch_time.timestamp()})

        rows.sort(key=lambda r: (r['stack_name'], r['version'], r['instance_id']))

        with OutputFormat(output):
            print_table(('stack_name version resource_id instance_id public_ip ' +
                         'private_ip state lb_status{} launch_time'.format(opt_docker_column)).split(),
                        rows, styles=STYLES, titles=TITLES)

        if piu is not None:
            odd_host = odd_host or Piu.find_odd_host(region)
            auto_connect = len(rows) == 1
            for row in rows:
                if row['private_ip'] is not None:
                    Piu.request_access(instance=row['private_ip'],
                                       reason=piu,
                                       odd_host=odd_host,
                                       connect=auto_connect)


@cli.command()
@click.argument('stack_ref', nargs=-1)
@region_option
@output_option
@watch_option
@watchrefresh_option
@stacktrace_visible_option
def status(stack_ref, region, output, w, watch):
    '''Show stack status information'''

    stack_refs = get_stack_refs(stack_ref)
    region = get_region(region)
    check_credentials(region)

    ec2 = boto3.resource('ec2', region)
    cf = boto3.resource('cloudformation', region)

    for _ in watching(w, watch):
        rows = []
        for stack in sorted(get_stacks(stack_refs, region)):
            instance_health = get_instance_health(stack.StackName, region)

            main_dns_resolves = None
            version_addresses = set()
            main_addresses = set()
            http_status = None
            for res in cf.Stack(stack.StackId).resource_summaries.all():
                if res.resource_type == 'AWS::Route53::RecordSet':
                    name = res.physical_resource_id
                    if not name:
                        # physical resource ID will be empty during stack creation
                        continue
                    if 'version' in res.logical_id.lower():
                        # VersionDomain -> check HTTPS reachability
                        # (won't work for internal ELBs, but we don't care here)
                        try:
                            requests.get('https://{}/'.format(name), timeout=1)
                            http_status = 'OK'
                        except:
                            http_status = 'ERROR'
                        version_addresses = resolve_to_ip_addresses(name)
                    else:
                        # MainDomain -> check whether DNS resolves to this stack version
                        main_addresses = resolve_to_ip_addresses(name)

            if version_addresses and main_addresses:
                main_dns_resolves = bool(version_addresses & main_addresses)

            instances = list(ec2.instances.filter(Filters=[{'Name': 'tag:aws:cloudformation:stack-id',
                                                            'Values': [stack.StackId]}]))
            rows.append({'stack_name': stack.name,
                         'version': stack.version,
                         'status': stack.StackStatus,
                         'total_instances': len(instances),
                         'running_instances': len([i for i in instances if i.state['Name'] == 'running']),
                         'healthy_instances': get_healthy_instances(instance_health),
                         'lb_status': ','.join(set(instance_health.values())),
                         'main_dns': main_dns_resolves,
                         'http_status': http_status
                         })

        with OutputFormat(output):
            print_table(('stack_name version status total_instances running_instances healthy_instances ' +
                         'lb_status http_status main_dns').split(), rows, styles=STYLES, titles=TITLES)


@cli.command()
@click.argument('stack_ref', nargs=-1)
@region_option
@output_option
@watch_option
@watchrefresh_option
@stacktrace_visible_option
def domains(stack_ref, region, output, w, watch):
    '''List the stack's Route53 domains'''

    stack_refs = get_stack_refs(stack_ref)
    region = get_region(region)
    check_credentials(region)

    cf = boto3.resource('cloudformation', region)

    records_by_name = {}

    for _ in watching(w, watch):
        rows = []
        for stack in get_stacks(stack_refs, region):
            if stack.StackStatus == 'ROLLBACK_COMPLETE':
                # performance optimization: do not call EC2 API for "dead" stacks
                continue

            for res in cf.Stack(stack.StackId).resource_summaries.all():
                if res.resource_type == 'AWS::Route53::RecordSet':
                    name = res.physical_resource_id
                    if name not in records_by_name:
                        hosted_zone = next(Route53.get_hosted_zones(name))
                        zone_name = hosted_zone.domain_name
                        for rec in get_records(zone_name):
                            record = Route53Record.from_boto_dict(rec)
                            records_by_name[(rec['Name'].rstrip('.'), rec.get('SetIdentifier'))] = record
                    record = (records_by_name.get((name, stack.StackName)) or
                              records_by_name.get((name, None)))  # type: Route53Record
                    row = {'stack_name': stack.name,
                           'version': stack.version,
                           'resource_id': res.logical_id,
                           'domain': res.physical_resource_id,
                           'weight': None,
                           'type': None,
                           'value': None,
                           'create_time': calendar.timegm(res.last_updated_timestamp.timetuple())}
                    if record:
                        if record.resource_records:
                            value = ','.join([r['Value']
                                              for r in record.resource_records])
                        elif record.alias_target:
                            value = record.alias_target['DNSName']
                        else:
                            value = ''
                        row.update({'weight': (str(record.weight)
                                               if record.weight is not None
                                               else ''),
                                    'type': record.type.value,
                                    'value': value})
                    rows.append(row)

        with OutputFormat(output):
            print_table('stack_name version resource_id domain weight type value create_time'.split(),
                        rows, styles=STYLES, titles=TITLES)


@cli.command()
@click.argument('stack_name')
@click.argument('stack_version', required=False)
@click.argument('percentage', type=FloatRange(0, 100, clamp=True), required=False)
@region_option
@output_option
@stacktrace_visible_option
def traffic(stack_name, stack_version, percentage, region, output):
    '''Route traffic to a specific stack (weighted DNS record)'''

    stack_refs = get_stack_refs([stack_name, stack_version])
    region = get_region(region)
    check_credentials(region)

    with OutputFormat(output):
        for ref in stack_refs:
            if percentage is None:
                print_version_traffic(ref, region)
            else:
                change_version_traffic(ref, percentage, region)


@cli.command()
@click.argument('stack_ref', nargs=-1)
@click.option('--hide-older-than',
              help='Hide images older than X days (default: 21)',
              type=int, default=21, metavar='DAYS')
@click.option('--show-instances', is_flag=True, help='Show EC2 instance IDs')
@region_option
@output_option
@stacktrace_visible_option
def images(stack_ref, region, output, hide_older_than, show_instances):
    '''Show all used AMIs and available Taupage AMIs'''

    stack_refs = get_stack_refs(stack_ref)
    region = get_region(region)
    check_credentials(region)

    ec2 = boto3.resource('ec2', region)

    instances_by_image = collections.defaultdict(list)
    for inst in ec2.instances.all():
        if inst.state['Name'] == 'terminated':
            # do not count TERMINATED EC2 instances
            continue
        stack_name = get_tag(inst.tags, 'aws:cloudformation:stack-name')
        if not stack_refs or matches_any(stack_name, stack_refs):
            instances_by_image[inst.image_id].append(inst)

    images = {}
    for image in ec2.images.filter(ImageIds=list(instances_by_image.keys())):
        images[image.id] = image
    if not stack_refs:
        filters = [{'Name': 'name', 'Values': ['*Taupage-*']},
                   {'Name': 'state', 'Values': ['available']}]
        for image in ec2.images.filter(Filters=filters):
            images[image.id] = image
    rows = []
    cutoff = datetime.datetime.now() - datetime.timedelta(days=hide_older_than)
    for image in images.values():
        row = image.meta.data.copy()
        creation_time = parse_time(image.creation_date)
        row['creation_time'] = creation_time
        row['instances'] = ', '.join(sorted(i.id for i in instances_by_image[image.id]))
        row['total_instances'] = len(instances_by_image[image.id])
        stacks = set()
        for instance in instances_by_image[image.id]:
            stack_name = get_tag(instance.tags, 'aws:cloudformation:stack-name')
            # EC2 instance might not be part of a CF stack
            if stack_name:
                stacks.add(stack_name)
        row['stacks'] = ', '.join(sorted(stacks))

        if creation_time is None:
            continue

        if creation_time > cutoff.timestamp() or row['total_instances']:
            rows.append(row)

    rows.sort(key=lambda x: x.get('Name'))
    with OutputFormat(output):
        cols = 'ImageId Name OwnerId Description stacks total_instances creation_time'
        if show_instances:
            cols = cols.replace('total_instances', 'instances')
        print_table(cols.split(), rows, titles=TITLES, max_column_widths=MAX_COLUMN_WIDTHS)


def is_ip_address(x: str):
    """
    Checks if x is a valid ip address.
    """
    try:
        ipaddress.ip_address(x)
        return True
    except ValueError:
        return False


def get_console_line_style(line: str):
    if 'ERROR:' in line:
        return {'fg': 'red', 'bold': True}
    elif 'WARNING:' in line:
        return {'fg': 'yellow', 'bold': True}
    elif 'SUCCESS:' in line:
        return {'fg': 'green', 'bold': True}
    elif 'INFO:' in line:
        return {'bold': True}
    else:
        return {}


def print_console(line: str):
    style = get_console_line_style(line)
    click.secho(line, **style)


@cli.command()
@click.argument('instance_or_stack_ref', nargs=-1)
@click.option('-l', '--limit', help='Show last N lines of console output (default: 25)',
              type=int, default=25, metavar='N')
@region_option
@watch_option
@watchrefresh_option
@stacktrace_visible_option
def console(instance_or_stack_ref, limit, region, w, watch):
    '''Print EC2 instance console output.

    INSTANCE_OR_STACK_REF can be an instance ID, private IP address or stack name/version.

    Ex:\n
    senza console my-stack-template.yml version1\n
    senza console my-stack-template.yml\n
    senza console my-stack-name version1\n
    senza console my-stack-name\n
    senza console 172.23.22.240\n
    senza console i-7c3a3de5\n
    senza console
    '''

    if instance_or_stack_ref and all(x.startswith('i-') for x in instance_or_stack_ref):
        stack_refs = None
        filters = [{'Name': 'instance-id', 'Values': list(instance_or_stack_ref)}]
    elif instance_or_stack_ref and all(is_ip_address(x) for x in instance_or_stack_ref):
        stack_refs = None
        filters = [{'Name': 'private-ip-address', 'Values': list(instance_or_stack_ref)}]
    else:
        stack_refs = get_stack_refs(instance_or_stack_ref)
        # filter out instances not part of any stack
        filters = [{'Name': 'tag-key', 'Values': ['aws:cloudformation:stack-name']}]

    region = get_region(region)
    check_credentials(region)

    ec2 = boto3.resource('ec2', region)

    for _ in watching(w, watch):

        found_match = False
        for instance in ec2.instances.filter(Filters=filters):

            cf_stack_name = get_tag(instance.tags, 'aws:cloudformation:stack-name')
            if not stack_refs or matches_any(cf_stack_name, stack_refs):
                found_match = True
                output = {}
                try:
                    output = instance.console_output()
                except:
                    pass
                click.secho('Showing last {} lines of {}/{}..'.format(limit,
                                                                      cf_stack_name,
                                                                      instance.private_ip_address or instance.id),
                            bold=True)
                if isinstance(output, dict) and output.get('Output'):
                    for line in output['Output'].split('\n')[-limit:]:
                        print_console(line)
        if not found_match:
            if stack_refs is None:
                fatal_error("No EC2 instance with id '{}'.".format(instance_or_stack_ref[0]))
            elif instance_or_stack_ref:
                fatal_error("No stack matching '{}'{}.".format(
                    stack_refs[0].name,
                    " with version '{}'".format(stack_refs[0].version) if stack_refs[0].version is not None else ''))
            else:
                fatal_error("No EC2 instances in region '{}'.".format(region))


@cli.command()
@click.argument('stack_ref', nargs=-1)
@region_option
@json_output_option
@stacktrace_visible_option
def dump(stack_ref, region, output):
    '''Dump Cloud Formation template of existing stack'''

    stack_refs = get_stack_refs(stack_ref)
    region = get_region(region)
    check_credentials(region)

    cf = BotoClientProxy('cloudformation', region)

    for stack in get_stacks(stack_refs, region):
        data = cf.get_template(StackName=stack.StackName)['TemplateBody']
        cfjson = json.dumps(data, sort_keys=True, indent=4)
        print_json(cfjson, output)


def get_auto_scaling_groups(stack_refs, region):
    cf = BotoClientProxy('cloudformation', region)
    for stack in get_stacks(stack_refs, region):
        resources = cf.describe_stack_resources(StackName=stack.StackName)['StackResources']

        for resource in resources:
            if resource['ResourceType'] == 'AWS::AutoScaling::AutoScalingGroup':
                asg_name = resource['PhysicalResourceId']
                yield asg_name


@cli.command()
@click.argument('stack_ref', nargs=-1)
@region_option
@click.option('--image',
              metavar='AMI_ID_OR_LATEST',
              help='Use specified image (AMI ID or "latest")')
@click.option('--instance-type',
              metavar='INSTANCE_TYPE',
              help='Use specified EC2 instance type')
@click.option('--user-data', metavar='YAML',
              help='Patch properties in user data YAML')
@stacktrace_visible_option
def patch(stack_ref, region, image, instance_type, user_data):
    '''Patch specific properties of existing stack.

    Currently only supports patching ASG launch configurations.'''

    stack_refs = get_stack_refs(stack_ref)
    region = get_region(region)
    check_credentials(region)

    if image == 'latest':
        image = find_taupage_image(region).id

    properties = {'ImageId': image,
                  'InstanceType': instance_type,
                  'UserData': yaml.safe_load(user_data) if user_data else None}
    # remove empty values
    properties = {k: v for k, v in properties.items() if v}

    if not properties:
        raise click.UsageError('Nothing to patch. Please specify at least one patch option (e.g. "--image").')

    asg = BotoClientProxy('autoscaling', region)

    for asg_name in get_auto_scaling_groups(stack_refs, region):
        with Action('Patching Auto Scaling Group {}..'.format(asg_name)) as act:
            result = asg.describe_auto_scaling_groups(AutoScalingGroupNames=[asg_name])
            groups = result['AutoScalingGroups']
            for group in groups:
                if not patch_auto_scaling_group(group, region, properties):
                    act.ok('NO CHANGES')


@cli.command('respawn-instances')
@click.argument('stack_ref', nargs=-1)
@click.option('--inplace',
              is_flag=True, help='Perform inplace update, do not scale out')
@click.option('-f', '--force',
              is_flag=True,
              help='Force respawn even if Launch Configuration is unchanged')
@region_option
@stacktrace_visible_option
def respawn_instances(stack_ref, inplace, force, region):
    '''Replace all EC2 instances in Auto Scaling Group(s)

    Performs a rolling update to prevent downtimes.'''

    stack_refs = get_stack_refs(stack_ref)
    region = get_region(region)
    check_credentials(region)

    for asg_name in get_auto_scaling_groups(stack_refs, region):
        respawn_auto_scaling_group(asg_name, region, inplace=inplace, force=force)


@cli.command()
@click.argument('stack_ref', nargs=-1)
@click.argument('desired_capacity', type=click.IntRange(0, 100, clamp=True))
@region_option
@stacktrace_visible_option
def scale(stack_ref, region, desired_capacity):
    '''Scale Auto Scaling Group to desired capacity'''

    stack_refs = get_stack_refs(stack_ref)
    region = get_region(region)
    check_credentials(region)

    asg = BotoClientProxy('autoscaling', region)

    for asg_name in get_auto_scaling_groups(stack_refs, region):
        group = get_auto_scaling_group(asg, asg_name)
        current_capacity = group['DesiredCapacity']
        with Action('Scaling {} from {} to {} instances..'.format(
                    asg_name, current_capacity, desired_capacity)) as act:
            if current_capacity == desired_capacity:
                act.ok('NO CHANGES')
            else:
                kwargs = {}
                if desired_capacity < group['MinSize']:
                    kwargs['MinSize'] = desired_capacity
                if desired_capacity > group['MaxSize']:
                    kwargs['MaxSize'] = desired_capacity
                asg.update_auto_scaling_group(AutoScalingGroupName=asg_name,
                                              DesiredCapacity=desired_capacity,
                                              **kwargs)


def failure_event(event: dict):
    status = event.get('ResourceStatus')
    return bool(event.get('ResourceStatusReason') and ('FAIL' in status or 'ROLLBACK' in status))


@cli.command()
@click.argument('stack_ref', nargs=-1)
@click.option('-d', '--deletion',
              is_flag=True, help='Wait for deletion instead of CREATE_COMPLETE')
@click.option('-t', '--timeout',
              type=click.IntRange(0, 7200, clamp=True),
              metavar='SECS',
              default=1800,
              help='Maximum wait time (default: 1800s)')
@click.option('-i', '--interval', default=5,
              type=click.IntRange(1, 600, clamp=True),
              help='Time between checks (default: 5s)')
@region_option
@stacktrace_visible_option
def wait(stack_ref, region, deletion, timeout, interval):
    """
    Wait for successful stack creation or deletion.

    Supports waiting for more than one stack up to timeout seconds.
    """

    stack_refs = get_stack_refs(stack_ref)
    region = get_region(region)
    cf = BotoClientProxy('cloudformation', region)

    target_status = (['DELETE_COMPLETE'] if deletion
                     else ['CREATE_COMPLETE', 'UPDATE_COMPLETE'])

    cutoff = time.time() + timeout

    while time.time() < cutoff:
        stacks_ok = set()
        stacks_in_progress = set()
        successful_actions = set()

        stacks_found = list(get_stacks(stack_refs, region, all=True, unique_only=True))

        if not stacks_found:
            raise click.UsageError('No matching stack for "{}" found'.format(' '.join(stack_ref)))

        for stack in stacks_found:

            if stack.StackStatus in target_status:
                successful_action, _ = stack.StackStatus.split('_')
                successful_actions.add('{}d'.format(successful_action.lower()))
                stacks_ok.add((stack.name, stack.version))

            elif stack.StackStatus.endswith('_IN_PROGRESS'):
                stacks_in_progress.add((stack.name, stack.version, stack.StackStatus))

            else:  # _FAILED or ROLLBACK_COMPLETE or UPDATE_ROLLBACK_COMPLETE
                # output event messages for troubleshooting
                events = cf.describe_stack_events(StackName=stack.StackId)['StackEvents']
                for event in sorted(events, key=lambda x: x['Timestamp']):
                    if failure_event(event):
                        error('ERROR: {LogicalResourceId} {ResourceStatus}: '
                              '{ResourceStatusReason}'.format(**event))
                fatal_error('ERROR: Stack {}-{} has '
                            'status {}'.format(stack.name,
                                               stack.version,
                                               stack.StackStatus))

        if stacks_in_progress:
            waiting_for = ', '.join(['{}-{} ({})'.format(*x)
                                     for x in sorted(stacks_in_progress)])
            info('Waiting up to {:.0f} more secs '
                 'for stack{} {}..'.format(cutoff - time.time(),
                                           's' if len(stacks_in_progress) > 1 else '',
                                           waiting_for))
            time.sleep(interval)
            continue
        if stacks_ok:
            successful_stacks = ', '.join(['{}-{}'.format(*x)
                                           for x in sorted(stacks_ok)])
            actions = '/'.join(successful_actions)
            ok('OK: Stack(s) {} {} successfully.'.format(successful_stacks,
                                                         actions))
            return
    raise click.Abort()

cli.add_command(cmd_config)


def main():
    HandleExceptions(cli)()


if __name__ == "__main__":
    main()




from senza.components.weighted_dns_elastic_load_balancer import component_weighted_dns_elastic_load_balancer
from senza.components.elastic_load_balancer_v2 import component_elastic_load_balancer_v2


def component_weighted_dns_elastic_load_balancer_v2(definition, configuration, args, info, force, account_info):
    return component_weighted_dns_elastic_load_balancer(definition, configuration, args, info, force, account_info,
                                                        lb_component=component_elastic_load_balancer_v2)



from subprocess import call
from typing import Optional

from ..exceptions import PiuNotFound
from ..manaus.route53 import Route53, Route53Record  # NOQA


class Piu:
    """
    Wrapper around `piu <https://github.com/zalando-stups/piu>`_

    For more information about `piu` see
    http://stups.readthedocs.io/en/latest/user-guide/ssh-access.html#ssh-access
    """
    @staticmethod
    def request_access(instance: str, reason: str, odd_host: Optional[str],
                       connect: bool):
        """
        Request SSH access to a single host
        """
        reason = '{} via senza'.format(reason)
        cmd = ['piu', 'request-access',
               instance, reason]

        if connect:
            cmd.append('--connect')

        if odd_host is not None:
            cmd.extend(['-O', odd_host])

        try:
            call(cmd)
        except FileNotFoundError:
            raise PiuNotFound

    @staticmethod
    def find_odd_host(region: str) -> Optional[str]:
        """
        Tries to find the odd host based on the region and route53 records
        """
        route53 = Route53()
        hosted_zones = list(route53.get_hosted_zones())
        for hosted_zone in hosted_zones:
            potential_name = 'odd-{region}.{domain}'.format(region=region,
                                                            domain=hosted_zone.name)
            records = route53.get_records(name=potential_name)
            try:
                record = next(records)  # type: Route53Record
            except StopIteration:
                # The domain name was not found
                pass
            else:
                odd_host = record.name[:-1]  # remove the trailing dot
                return odd_host

        return None



from collections import OrderedDict
from typing import Dict, Iterator, List, Optional

import boto3

from .exceptions import VPCError


class EC2VPC:

    """
    See:
    http://boto3.readthedocs.io/en/latest/reference/services/ec2.html#vpc
    """

    def __init__(self,
                 vpc_id: str,
                 is_default: bool,
                 tags: Optional[List[Dict[str, str]]]):
        self.vpc_id = vpc_id
        self.is_default = is_default
        tags = tags or []  # type: List[Dict[str, str]]
        self.tags = OrderedDict([(t['Key'], t['Value']) for t in tags])  # type: Dict[str, str]

        self.name = self.tags.get('Name', self.vpc_id)

    def __str__(self):
        return '{name} ({vpc_id})'.format_map(vars(self))

    def __repr__(self):
        return '<EC2VPC: {name} ({vpc_id})>'.format_map(vars(self))

    @classmethod
    def from_boto_vpc(cls, vpc) -> "EC2VPC":
        """
        Converts an ec2.VPC as returned by resource.vpcs.all()

        See:
        http://boto3.readthedocs.io/en/latest/reference/services/ec2.html#vpc
        """

        return cls(vpc.vpc_id, vpc.is_default, vpc.tags)


class EC2:

    def __init__(self, region: str):
        self.region = region

    def get_all_vpcs(self) -> Iterator[EC2VPC]:
        """
        Get all VPCs from the account
        """
        resource = boto3.resource('ec2', self.region)

        for vpc in resource.vpcs.all():
            yield EC2VPC.from_boto_vpc(vpc)

    def get_default_vpc(self) -> EC2VPC:
        """
        Get one VPC from the account, either the default or, if only one
        exists, that one.
        """
        resource = boto3.resource('ec2', self.region)

        number_of_vpcs = 0
        # We shouldn't use the list with .all() because it has internal paging!
        for vpc_number, vpc in enumerate(resource.vpcs.all(), start=1):
            number_of_vpcs = vpc_number

            if vpc.is_default:
                return EC2VPC.from_boto_vpc(vpc)

            if vpc_number == 1:
                first_vpc = vpc

        if number_of_vpcs == 0:
            raise VPCError("Can't find any VPC!", number_of_vpcs)
        elif number_of_vpcs == 1:
            # Use the only one VPC if it's not the default VPC found
            return EC2VPC.from_boto_vpc(first_vpc)
        else:
            raise VPCError("Multiple VPCs are only supported if one "
                           "VPC is the default VPC (IsDefault=true)!",
                           number_of_vpcs)



"""
Random functions that are useful in several places and not don't fall under
the domain of any other module.
"""

import re
import pystache


def named_value(dictionary):
    """
    Gets the name and value of a dict with a single key (for example SenzaInfo
    parameters or Senza Components)
    """
    return next(iter(dictionary.items()))


def ensure_keys(dict_obj, *keys):
    """
    Ensure ``dict_obj`` has the hierarchy ``{keys[0]: {keys[1]: {...}}}``

    The innermost key will have ``{}`` has value if didn't exist already.
    """
    if len(keys) == 0:
        return dict_obj
    else:
        first, rest = keys[0], keys[1:]
        if first not in dict_obj:
            dict_obj[first] = {}
        dict_obj[first] = ensure_keys(dict_obj[first], *rest)
        return dict_obj


def camel_case_to_underscore(name):
    """
    Converts name from CamelCase to snake_case
    """
    # the two steps are needed to support words with sequences of more than
    # one uppercase character
    step1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', step1).lower()


def pystache_render(*args, **kwargs):
    """
    Render pystache template with strict mode
    """
    render = pystache.Renderer(missing_tags='strict')
    return render.render(*args, **kwargs)



class SenzaException(Exception):
    """
    Base class for Senza exceptions
    """


class InvalidState(SenzaException):
    """
    Exception raised when executing an action would try to change a stack
    to an invalid state
    """


class PiuNotFound(SenzaException, FileNotFoundError):
    """
    Error raised when piu executable is not found
    """

    def __init__(self):
        super().__init__('Command not found: piu')


class InvalidConfigKey(SenzaException, ValueError):
    """
    Error raised when trying to use an Invalid Config Key
    """

    def __init__(self, message: str):
        super().__init__(message)


class InvalidDefinition(SenzaException):
    """
    Exception raised when trying to parse and invalid senza definition
    """

    def __init__(self, path: str, reason: str):
        self.path = path
        self.reason = reason

    def __str__(self):
        return ("{path} is not a valid senza definition: "
                "{reason}".format_map(vars(self)))



'''
HA Postgres app, which needs an S3 bucket to store WAL files
'''

import random
import string

import click
import requests
from clickclick import choice, warning
from senza.aws import (encrypt, get_security_group, get_vpc_attribute,
                       list_kms_keys)
from senza.utils import pystache_render

from ..manaus.boto_proxy import BotoClientProxy
from ._helper import check_s3_bucket, get_account_alias, prompt

POSTGRES_PORT = 5432
HEALTHCHECK_PORT = 8008

# This template goes through 2 formatting phases. Once during the init phase and once during
# the create phase of senza. Some placeholders should be evaluated during create.
# This creates some ugly placeholder formatting, therefore some placeholders are placeholders for placeholders
# - version
# - ImageVersion
TEMPLATE = '''
# basic information for generating and executing this definition
SenzaInfo:
  StackName: spilo
  {{^docker_image}}
  Parameters:
    - ImageVersion:
        Description: "Docker image version of spilo."
  {{/docker_image}}
  Tags:
    - SpiloCluster: "{{version}}"

# a list of senza components to apply to the definition
SenzaComponents:

  # this basic configuration is required for the other components
  - Configuration:
      Type: Senza::StupsAutoConfiguration # auto-detect network setup

  # will create a launch configuration and auto scaling group with scaling triggers
  - AppServer:
      Type: Senza::TaupageAutoScalingGroup
      AutoScaling:
        Minimum: 3
        Maximum: 3
        MetricType: CPU
      InstanceType: {{instance_type}}
      {{#ebs_optimized}}
      EbsOptimized: True
      {{/ebs_optimized}}
      BlockDeviceMappings:
        - DeviceName: /dev/xvdk
          {{#use_ebs}}
          Ebs:
            VolumeSize: {{volume_size}}
            VolumeType: {{volume_type}}
            {{#snapshot_id}}
            SnapshotId: {{snapshot_id}}
            {{/snapshot_id}}
            {{#volume_iops}}
            Iops: {{volume_iops}}
            {{/volume_iops}}
          {{/use_ebs}}
          {{^use_ebs}}
          VirtualName: ephemeral0
          {{/use_ebs}}
      ElasticLoadBalancer:
        - PostgresLoadBalancer
        {{#add_replica_loadbalancer}}
        - PostgresReplicaLoadBalancer
        {{/add_replica_loadbalancer}}
      HealthCheckType: EC2
      SecurityGroups:
        - Fn::GetAtt:
          - SpiloMemberSG
          - GroupId
      IamRoles:
        - Ref: PostgresAccessRole
      AssociatePublicIpAddress: false # change for standalone deployment in default VPC
      TaupageConfig:
        runtime: Docker
        {{#docker_image}}
        source: {{docker_image}}
        {{/docker_image}}
        {{^docker_image}}
        source: "{{ImageVersion}}"
        {{/docker_image}}
        ports:
          {{postgres_port}}: {{postgres_port}}
          {{healthcheck_port}}: {{healthcheck_port}}
        etcd_discovery_domain: "{{discovery_domain}}"
        environment:
          SCOPE: "{{version}}"
          ETCD_DISCOVERY_DOMAIN: "{{discovery_domain}}"
          WAL_S3_BUCKET: "{{wal_s3_bucket}}"
          PGPASSWORD_SUPERUSER: "{{pgpassword_superuser}}"
          PGPASSWORD_ADMIN: "{{pgpassword_admin}}"
          PGPASSWORD_STANDBY: "{{pgpassword_standby}}"
          BACKUP_SCHEDULE: "00 01 * * *"
          PATRONI_CONFIGURATION: | ## https://github.com/zalando/patroni#yaml-configuration
            bootstrap:
              dcs:
                postgresql:
                  parameters:
                    log_connections: on
              pg_hba:
               - hostnossl all all all reject
        root: True
        sysctl:
          vm.overcommit_memory: 2
          vm.overcommit_ratio: 200
          vm.dirty_ratio: 8
          vm.dirty_background_ratio: 1
          vm.swappiness: 1
        appdynamics_application: "postgresapp-{{version}}"
        mounts:
          /home/postgres/pgdata:
            partition: /dev/xvdk
            filesystem: {{fstype}}
            {{#snapshot_id}}
            erase_on_boot: false
            {{/snapshot_id}}
            {{^snapshot_id}}
            erase_on_boot: true
            {{/snapshot_id}}
            options: {{fsoptions}}
Resources:
  {{#add_replica_loadbalancer}}
  PostgresReplicaRoute53Record:
    Type: AWS::Route53::RecordSet
    Properties:
      Type: CNAME
      TTL: 20
      HostedZoneName: {{hosted_zone}}
      Name: "{{version}}-replica.{{hosted_zone}}"
      ResourceRecords:
        - Fn::GetAtt:
           - PostgresReplicaLoadBalancer
           - DNSName
  PostgresReplicaLoadBalancer:
    Type: AWS::ElasticLoadBalancing::LoadBalancer
    Properties:
      CrossZone: true
      HealthCheck:
        HealthyThreshold: 2
        Interval: 5
        Target: HTTP:{{healthcheck_port}}/replica
        Timeout: 3
        UnhealthyThreshold: 2
      Listeners:
        - InstancePort: {{postgres_port}}
          LoadBalancerPort: {{postgres_port}}
          Protocol: TCP
      LoadBalancerName: "spilo-{{version}}-replica"
      ConnectionSettings:
        IdleTimeout: 3600
      SecurityGroups:
        - Fn::GetAtt:
          - SpiloReplicaSG
          - GroupId
      Scheme: internal
      Subnets:
        Fn::FindInMap:
          - LoadBalancerSubnets
          - Ref: AWS::Region
          - Subnets
  {{/add_replica_loadbalancer}}
  PostgresRoute53Record:
    Type: AWS::Route53::RecordSet
    Properties:
      Type: CNAME
      TTL: 20
      HostedZoneName: {{hosted_zone}}
      Name: "{{version}}.{{hosted_zone}}"
      ResourceRecords:
        - Fn::GetAtt:
           - PostgresLoadBalancer
           - DNSName
  PostgresLoadBalancer:
    Type: AWS::ElasticLoadBalancing::LoadBalancer
    Properties:
      CrossZone: true
      HealthCheck:
        HealthyThreshold: 2
        Interval: 5
        Target: HTTP:{{healthcheck_port}}/master
        Timeout: 3
        UnhealthyThreshold: 2
      Listeners:
        - InstancePort: {{postgres_port}}
          LoadBalancerPort: {{postgres_port}}
          Protocol: TCP
      LoadBalancerName: "spilo-{{version}}"
      ConnectionSettings:
        IdleTimeout: 3600
      SecurityGroups:
        - Fn::GetAtt:
          - SpiloMasterSG
          - GroupId
      Scheme: internal
      Subnets:
        Fn::FindInMap:
          - LoadBalancerSubnets
          - Ref: AWS::Region
          - Subnets
  PostgresAccessRole:
    Type: AWS::IAM::Role
    Properties:
      AssumeRolePolicyDocument:
        Version: "2012-10-17"
        Statement:
        - Effect: Allow
          Principal:
            Service: ec2.amazonaws.com
          Action: sts:AssumeRole
      Path: /
      Policies:
      - PolicyName: SpiloEC2S3KMSAccess
        PolicyDocument:
          Version: "2012-10-17"
          Statement:
          - Effect: Allow
            Action:
              - s3:ListBucket
            Resource:
              - "arn:aws:s3:::{{wal_s3_bucket}}"
              - "arn:aws:s3:::{{wal_s3_bucket}}/*"
          - Effect: Allow
            Action:
              - s3:*
            Resource:
              - "arn:aws:s3:::{{wal_s3_bucket}}/spilo/{{version}}/*"
          - Effect: Allow
            Action: ec2:CreateTags
            Resource: "*"
          - Effect: Allow
            Action: ec2:Describe*
            Resource: "*"
          {{#kms_arn}}
          - Effect: Allow
            Action:
              - "kms:Decrypt"
            Resource:
              - {{kms_arn}}
          {{/kms_arn}}
  SpiloMasterSG:
    Type: "AWS::EC2::SecurityGroup"
    Properties:
      GroupDescription: "Security Group for the master ELB of Spilo: {{version}}"
      SecurityGroupIngress:
        - IpProtocol: tcp
          FromPort: {{postgres_port}}
          ToPort: {{postgres_port}}
          CidrIp: {{elb_access_cidr}}
  {{#add_replica_loadbalancer}}
  SpiloReplicaSG:
    Type: "AWS::EC2::SecurityGroup"
    Properties:
      GroupDescription: "Security Group for the replica ELB of Spilo: {{version}}"
      SecurityGroupIngress:
        - IpProtocol: tcp
          FromPort: {{postgres_port}}
          ToPort: {{postgres_port}}
          CidrIp: {{elb_access_cidr}}
  {{/add_replica_loadbalancer}}
  SpiloMemberSG:
    Type: "AWS::EC2::SecurityGroup"
    Properties:
      GroupDescription: "Security Group for members of Spilo: {{version}}"
      SecurityGroupIngress:
        - IpProtocol: tcp
          FromPort: {{postgres_port}}
          ToPort: {{postgres_port}}
          SourceSecurityGroupId:
            Fn::GetAtt:
              - SpiloMasterSG
              - GroupId
        - IpProtocol: tcp
          FromPort: {{healthcheck_port}}
          ToPort: {{healthcheck_port}}
          SourceSecurityGroupId:
            Fn::GetAtt:
              - SpiloMasterSG
              - GroupId

        {{#add_replica_loadbalancer}}
        - IpProtocol: tcp
          FromPort: {{postgres_port}}
          ToPort: {{postgres_port}}
          SourceSecurityGroupId:
            Fn::GetAtt:
              - SpiloReplicaSG
              - GroupId
        - IpProtocol: tcp
          FromPort: {{healthcheck_port}}
          ToPort: {{healthcheck_port}}
          SourceSecurityGroupId:
            Fn::GetAtt:
              - SpiloReplicaSG
              - GroupId
        {{/add_replica_loadbalancer}}
        {{#zmon_sg_id}}
        - IpProtocol: tcp
          FromPort: {{promotheus_port}}
          ToPort: {{promotheus_port}}
          SourceSecurityGroupId: "{{zmon_sg_id}}"
        - IpProtocol: tcp
          FromPort: {{postgres_port}}
          ToPort: {{postgres_port}}
          SourceSecurityGroupId: "{{zmon_sg_id}}"
        - IpProtocol: tcp
          FromPort: {{healthcheck_port}}
          ToPort: {{healthcheck_port}}
          SourceSecurityGroupId: "{{zmon_sg_id}}"
        {{/zmon_sg_id}}
        {{#odd_sg_id}}
        - IpProtocol: tcp
          FromPort: 0
          ToPort: 65535
          SourceSecurityGroupId: "{{odd_sg_id}}"
        {{/odd_sg_id}}
  SpiloMemberIngressMembers:
    Type: "AWS::EC2::SecurityGroupIngress"
    Properties:
      GroupId:
        Fn::GetAtt:
          - SpiloMemberSG
          - GroupId
      IpProtocol: tcp
      FromPort: 0
      ToPort: 65535
      SourceSecurityGroupId:
        Fn::GetAtt:
          - SpiloMemberSG
          - GroupId
'''


def ebs_optimized_supported(instance_type):
    """
    Per http://docs.aws.amazon.com/AWSEC2/latest/UserGuide/EBSOptimized.html
    """
    # TODO move to manaus
    return instance_type in ('c1.large', 'c3.xlarge', 'c3.2xlarge', 'c3.4xlarge',
                             'c4.large', 'c4.xlarge', 'c4.2xlarge', 'c4.4xlarge', 'c4.8xlarge',
                             'd2.xlarge', 'd2.2xlarge', 'd2.4xlarge', 'd2.8xlarge',
                             'g2.2xlarge', 'i2.xlarge', 'i2.2xlarge', 'i2.4xlarge',
                             'm1.large', 'm1.xlarge', 'm2.2xlarge', 'm2.4xlarge',
                             'm3.xlarge', 'm3.2xlarge', 'r3.xlarge', 'r3.2xlarge',
                             'r3.4xlarge')


def set_default_variables(variables):
    variables.setdefault('version', '{{Arguments.version}}')
    variables.setdefault('ImageVersion', '{{Arguments.ImageVersion}}')
    variables.setdefault('discovery_domain', 'postgres.example.com')
    variables.setdefault('docker_image', None)
    variables.setdefault('ebs_optimized', None)
    variables.setdefault('elb_access_cidr', '0.0.0.0/0')
    variables.setdefault('fsoptions', 'noatime,nodiratime,nobarrier')
    variables.setdefault('fstype', 'ext4')
    variables.setdefault('healthcheck_port', HEALTHCHECK_PORT)
    variables.setdefault('hosted_zone', 'example.com')
    variables.setdefault('add_replica_loadbalancer', False)
    variables.setdefault('instance_type', 't2.medium')
    variables.setdefault('kms_arn', None)
    variables.setdefault('odd_sg_id', None)
    variables.setdefault('pgpassword_admin', 'admin')
    variables.setdefault('pgpassword_standby', 'standby')
    variables.setdefault('pgpassword_superuser', 'zalando')
    variables.setdefault('postgres_port', POSTGRES_PORT)
    variables.setdefault('promotheus_port', '9100')
    variables.setdefault('snapshot_id', None)
    variables.setdefault('use_ebs', True)
    variables.setdefault('volume_iops', 300)
    variables.setdefault('volume_size', 10)
    variables.setdefault('volume_type', 'gp2')
    variables.setdefault('wal_s3_bucket', None)
    variables.setdefault('zmon_sg_id', None)

    return variables


def gather_user_variables(variables, region, account_info):
    defaults = set_default_variables(dict())

    if click.confirm('Do you want to set the docker image now? [No]'):
        prompt(variables, "docker_image", "Docker Image Version", default=get_latest_image())

    prompt(variables, 'wal_s3_bucket', 'Postgres WAL S3 bucket to use',
           default='{}-{}-spilo-app'.format(get_account_alias(), region))

    prompt(variables, 'instance_type', 'EC2 instance type', default='t2.medium')

    variables['hosted_zone'] = account_info.Domain or defaults['hosted_zone']
    if (variables['hosted_zone'][-1:] != '.'):
        variables['hosted_zone'] += '.'
    prompt(variables, 'discovery_domain', 'ETCD Discovery Domain',
           default='postgres.' + variables['hosted_zone'][:-1])

    variables['add_replica_loadbalancer'] = click.confirm('Do you want a replica ELB?', default=False)

    prompt(variables, 'elb_access_cidr', 'Which network should be allowed to access the ELB''s? (default=vpc)',
           default=get_vpc_attribute(region=region, vpc_id=account_info.VpcID, attribute='cidr_block'))

    odd_sg_name = 'Odd (SSH Bastion Host)'
    odd_sg = get_security_group(region, odd_sg_name)
    if odd_sg and click.confirm('Do you want to allow access to the Spilo nodes from {}?'.format(odd_sg_name),
                                default=True):
        variables['odd_sg_id'] = odd_sg.group_id

    # Find all Security Groups attached to the zmon worker with 'zmon' in their name
    ec2 = BotoClientProxy('ec2', region)
    filters = [{'Name': 'tag-key', 'Values': ['StackName']}, {'Name': 'tag-value', 'Values': ['zmon-appliance']}]
    zmon_sgs = list()
    for reservation in ec2.describe_instances(Filters=filters).get('Reservations', []):
        for instance in reservation.get('Instances', []):
            zmon_sgs += [sg['GroupId'] for sg in instance.get('SecurityGroups', []) if 'zmon' in sg['GroupName']]

    if len(zmon_sgs) == 0:
        warning('Could not find zmon security group, do you have the zmon-appliance deployed?')
    else:
        click.confirm('Do you want to allow access to the Spilo nodes from zmon?', default=True)
        if len(zmon_sgs) > 1:
            prompt(variables, 'zmon_sg_id', 'Which Security Group should we allow access from? {}'.format(zmon_sgs))
        else:
            variables['zmon_sg_id'] = zmon_sgs[0]

    if variables['instance_type'].lower().split('.')[0] in ('c3', 'g2', 'hi1', 'i2', 'm3', 'r3'):
        variables['use_ebs'] = click.confirm('Do you want database data directory on external (EBS) storage? [Yes]',
                                             default=defaults['use_ebs'])
    else:
        variables['use_ebs'] = True

    if variables['use_ebs']:
        prompt(variables, 'volume_size', 'Database volume size (GB, 10 or more)', default=defaults['volume_size'])
        prompt(variables, 'volume_type', 'Database volume type (gp2, io1 or standard)',
               default=defaults['volume_type'])
        if variables['volume_type'] == 'io1':
            pio_max = variables['volume_size'] * 30
            prompt(variables, "volume_iops", 'Provisioned I/O operations per second (100 - {0})'.
                   format(pio_max), default=str(pio_max))
        prompt(variables, "snapshot_id", "ID of the snapshot to populate EBS volume from", default="")
        if ebs_optimized_supported(variables['instance_type']):
            variables['ebs_optimized'] = True
    prompt(variables, "fstype", "Filesystem for the data partition", default=defaults['fstype'])
    prompt(variables, "fsoptions", "Filesystem mount options (comma-separated)",
           default=defaults['fsoptions'])

    prompt(variables, 'pgpassword_superuser', "Password for PostgreSQL superuser [random]", show_default=False,
           default=generate_random_password, hide_input=True, confirmation_prompt=True)
    prompt(variables, 'pgpassword_standby', "Password for PostgreSQL user standby [random]", show_default=False,
           default=generate_random_password, hide_input=True, confirmation_prompt=True)
    prompt(variables, 'pgpassword_admin', "Password for PostgreSQL user admin", show_default=True,
           default=defaults['pgpassword_admin'], hide_input=True, confirmation_prompt=True)

    if click.confirm('Do you wish to encrypt these passwords using KMS?', default=False):
        kms_keys = [k for k in list_kms_keys(region) if 'alias/aws/ebs' not in k['aliases']]

        if len(kms_keys) == 0:
            raise click.UsageError('No KMS key is available for encrypting and decrypting. '
                                   'Ensure you have at least 1 key available.')

        options = ['{}: {}'.format(k['KeyId'], k['Description']) for k in kms_keys]
        kms_key = choice(prompt='Please select the encryption key', options=options)
        kms_keyid = kms_key.split(':')[0]

        variables['kms_arn'] = [k['Arn'] for k in kms_keys if k['KeyId'] == kms_keyid][0]

        for key in [k for k in variables if k.startswith('pgpassword_')]:
            if variables[key]:
                encrypted = encrypt(region=region, KeyId=kms_keyid, Plaintext=variables[key], b64encode=True)
                variables[key] = 'aws:kms:{}'.format(encrypted)

    set_default_variables(variables)

    check_s3_bucket(variables['wal_s3_bucket'], region)

    return variables


def generate_random_password(length=64):
    """
    Generates a random password containing upper case characters and digits
    """
    char_list = string.ascii_uppercase + string.digits
    return ''.join(random.SystemRandom().choice(char_list)
                   for _ in range(length))


def generate_definition(variables):
    definition_yaml = pystache_render(TEMPLATE, variables)
    return definition_yaml


def get_latest_image(registry_domain='registry.opensource.zalan.do', team='acid', artifact='spilo-9.5'):
    """
    Gets the full name of latest image for an artifact
    """
    try:
        r = requests.get('https://{0}/teams/{1}/artifacts/{2}/tags'.format(registry_domain, team, artifact))
        if r.ok:
            # sort the tags by creation date
            latest = None
            for entry in sorted(r.json(), key=lambda t: t['created'], reverse=True):
                tag = entry['name']
                # try to avoid snapshots if possible
                if 'SNAPSHOT' not in tag:
                    latest = tag
                    break
                latest = latest or tag
            return "{0}/{1}/{2}:{3}".format(registry_domain, team, artifact, latest)
    except:
        pass
    return ""



from typing import Dict, Optional  # noqa: F401

from botocore.exceptions import ClientError

__all__ = ["extract_client_error_code"]


def extract_client_error_code(exception: ClientError) -> Optional[str]:
    """
    Extracts the client error code from a boto ClientError exception. Returns
    None if it fails.
    """
    error = exception.response.get('Error', {})  # type: Dict[str, Optional[str]]
    error_code = error.get('Code')
    return error_code



from datetime import datetime, timezone
from typing import Any, Dict, Iterator, Optional, Union

import boto3
from botocore.exceptions import ClientError

from .boto_proxy import BotoClientProxy


class IAMServerCertificate:
    """
    Server certificate stored in IAM.

    See:
    http://boto3.readthedocs.io/en/latest/reference/services/iam.html#IAM.Client.get_server_certificate
    """

    def __init__(self,
                 metadata: Dict[str, Union[str, datetime]],
                 certificate_body: str,
                 certificate_chain: str):

        self.metadata = metadata
        self.certificate_body = certificate_body
        self.certificate_chain = certificate_chain

        # metadata properties
        self.name = metadata['ServerCertificateName']  # type: str
        self.arn = metadata['Arn']  # type: str
        self.expiration = metadata['Expiration']  # type: datetime
        self.path = metadata['Path']  # type: str
        self.certificate_id = metadata['ServerCertificateId']  # type: str
        self.upload_date = metadata['UploadDate']  # type: datetime

    def __lt__(self, other: "IAMServerCertificate"):
        return self.upload_date < other.upload_date

    def __eq__(self, other: "IAMServerCertificate"):
        return self.arn == other.arn

    def __repr__(self):
        return "<IAMServerCertificate: {name}>".format_map(vars(self))

    @classmethod
    def from_boto_dict(cls,
                       server_certificate: Dict[str, Any]) -> "IAMServerCertificate":

        metadata = server_certificate['ServerCertificateMetadata']
        certificate_body = server_certificate['CertificateBody']
        certificate_chain = server_certificate['CertificateChain']

        return cls(metadata, certificate_body, certificate_chain)

    @classmethod
    def from_boto_server_certificate(cls, server_certificate) -> "IAMServerCertificate":
        """
        Converts an ServerCertificate as returned by server_certificates.all()
        """
        metadata = server_certificate.server_certificate_metadata
        certificate_body = server_certificate.certificate_body
        certificate_chain = server_certificate.certificate_chain

        return cls(metadata, certificate_body, certificate_chain)

    @classmethod
    def get_by_name(cls, region: str, name: str) -> "IAMServerCertificate":
        """
        Get IAMServerCertificate using the name of the server certificate
        """
        client = BotoClientProxy('iam', region)
        iam = IAM(region)

        try:
            response = client.get_server_certificate(ServerCertificateName=name)
            server_certificate = response['ServerCertificate']
            certificate = cls.from_boto_dict(server_certificate)
        except ClientError as error:
            # IAM.get_certificates can get certificates with a suffix
            certificates = sorted(iam.get_certificates(name=name),
                                  reverse=True)
            try:
                # try to return the latest certificate that matches the name
                certificate = certificates[0]
            except IndexError:
                raise error

        return certificate

    @staticmethod
    def arn_is_server_certificate(arn: Optional[str] = None):
        """
        Checks if the Amazon Resource Name (ARN) refers to an iam
        server certificate.

        See:
        http://docs.aws.amazon.com/IAM/latest/UserGuide/reference_identifiers.html
        """
        if arn is None:
            return False
        else:
            return (arn.startswith("arn:aws:iam:") and
                    'server-certificate' in arn)

    def is_valid(self, when: Optional[datetime]=None) -> bool:
        """
        Checks if the certificate is still valid
        """
        when = when if when is not None else datetime.now(timezone.utc)

        return when < self.expiration


class IAM:

    def __init__(self, region: str):
        self.region = region

    def get_certificates(self,
                         *,
                         valid_only: bool=True,
                         name: Optional[str] = None) -> Iterator[IAMServerCertificate]:
        resource = boto3.resource('iam', self.region)

        for server_certificate in resource.server_certificates.all():
            certificate = IAMServerCertificate.from_boto_server_certificate(server_certificate)

            if name is not None and not certificate.name.startswith(name):
                continue

            if valid_only and not certificate.is_valid():
                continue

            yield certificate



from datetime import datetime, timezone
from enum import Enum
from functools import total_ordering
from ssl import CertificateError, match_hostname
from typing import Any, Dict, Iterator, List, Optional

from .boto_proxy import BotoClientProxy


class ACMCertificateStatus(str, Enum):
    expired = "EXPIRED"
    failed = "FAILED"
    inactive = "INACTIVE"
    issued = "ISSUED"
    pending_validation = "PENDING_VALIDATION"
    revoked = "REVOKED"
    validation_timed_out = "VALIDATION_TIMED_OUT"


@total_ordering
class ACMCertificate:
    """
    See:
    http://boto3.readthedocs.io/en/latest/reference/services/acm.html#ACM.Client.list_certificates
    http://boto3.readthedocs.io/en/latest/reference/services/acm.html#ACM.Client.describe_certificate
    """
    def __init__(self,
                 domain_name: str,
                 arn: str,
                 subject_alternative_name: List[str],
                 domain_validation_options: List[Dict],
                 serial: str,
                 subject: str,
                 issuer: str,
                 created_at: datetime,
                 issued_at: datetime,
                 status: str,
                 not_before: datetime,
                 not_after: datetime,
                 signature_algorithm: str,
                 in_use_by: List[str],
                 revoked_at: Optional[datetime],
                 revocation_reason: Optional[str]):
        self.domain_name = domain_name
        self.arn = arn
        self.subject_alternative_name = subject_alternative_name
        self.domain_validation_options = domain_validation_options
        self.serial = serial
        self.subject = subject
        self.issuer = issuer
        self.created_at = created_at
        self.issued_at = issued_at
        self.status = ACMCertificateStatus(status)
        self.not_before = not_before
        self.not_after = not_after
        self.signature_algorithm = signature_algorithm
        self.in_use_by = in_use_by

        self.revoked_at = revoked_at
        self.revocation_reason = revocation_reason

    def __lt__(self, other: "ACMCertificate"):
        return self.created_at < other.created_at

    def __eq__(self, other: "ACMCertificate"):
        return self.arn == other.arn

    def __repr__(self):
        return "<ACMCertificate:{domain_name} ({arn})>".format_map(vars(self))

    @classmethod
    def from_boto_dict(cls,
                       certificate: Dict[str, Any]) -> "ACMCertificate":
        """
        Creates an ACMCertificate based on the dictionary returned by
        describe_certificate
        """

        domain_name = certificate['DomainName']
        arn = certificate['CertificateArn']
        subject_alternative_name = certificate['SubjectAlternativeNames']
        domain_validation_options = certificate['DomainValidationOptions']
        subject = certificate['Subject']
        created_at = certificate['CreatedAt']
        status = certificate['Status']
        signature_algorithm = certificate['SignatureAlgorithm']
        in_use_by = certificate['InUseBy']
        serial = certificate.get('Serial')
        issuer = certificate.get('Issuer')
        issued_at = certificate.get('IssuedAt')
        not_before = certificate.get('NotBefore')
        not_after = certificate.get('NotAfter')

        revoked_at = certificate.get('RevokedAt')
        revocation_reason = certificate.get('RevocationReason')

        return cls(domain_name, arn, subject_alternative_name,
                   domain_validation_options, serial, subject, issuer,
                   created_at, issued_at, status, not_before, not_after,
                   signature_algorithm, in_use_by,
                   revoked_at, revocation_reason)

    @classmethod
    def get_by_arn(cls, region: str, arn: str) -> "ACMCertificate":
        """
        Gets a ACMCertificate based on ARN alone
        """
        client = BotoClientProxy('acm', region)
        certificate = client.describe_certificate(CertificateArn=arn)['Certificate']
        return cls.from_boto_dict(certificate)

    @staticmethod
    def arn_is_acm_certificate(arn: Optional[str]=None) -> bool:
        if arn is None:
            return False
        else:
            return arn.startswith("arn:aws:acm:")

    def is_valid(self, when: Optional[datetime]=None) -> bool:
        """
        Checks if the certificate is still valid
        """
        when = when if when is not None else datetime.now(timezone.utc)

        if self.status != ACMCertificateStatus.issued:
            return False

        return self.not_before < when < self.not_after

    def matches(self, domain_name: str) -> bool:
        """
        Checks if certificate subject or alt names match the domain name.
        """
        # python ssl friendly certificate:
        subject = ((('commonName', self.domain_name),),)
        alt_name = [('DNS', name) for name in self.subject_alternative_name]
        certificate = {'subject': subject,
                       'subjectAltName': alt_name}

        try:
            match_hostname(certificate, domain_name)
        except CertificateError:
            return False
        else:
            return True


class ACM:
    """
    From https://aws.amazon.com/certificate-manager/

    AWS Certificate Manager is a service that lets you easily provision,
    manage, and deploy Secure Sockets Layer/Transport Layer Security (SSL/TLS)
    certificates for use with AWS services.

    See http://boto3.readthedocs.io/en/latest/reference/services/acm.html
    """

    def __init__(self, region=str):
        self.region = region

    def get_certificates(self,
                         *,
                         valid_only: bool=True,
                         domain_name: Optional[str]=None) -> Iterator[ACMCertificate]:
        """
        Gets certificates from ACM. By default it returns all valid certificates

        :param region: AWS region
        :param valid_only: Return only valid certificates
        :param domain_name: Return only certificates that match the domain
        """
        # TODO implement pagination
        client = BotoClientProxy('acm', self.region)
        certificates = client.list_certificates()['CertificateSummaryList']
        for summary in certificates:
            arn = summary['CertificateArn']
            certificate = ACMCertificate.get_by_arn(self.region, arn)
            if valid_only and not certificate.is_valid():
                pass
            elif domain_name is not None and not certificate.matches(domain_name):
                pass
            else:
                yield certificate



import click
from senza.aws import resolve_security_groups
from senza.components.elastic_load_balancer import ALLOWED_LOADBALANCER_SCHEMES, get_load_balancer_name, get_ssl_cert

from ..cli import AccountArguments, TemplateArguments
from ..manaus.route53 import convert_cname_records_to_alias

SENZA_PROPERTIES = frozenset(['Domains', 'HealthCheckPath', 'HealthCheckPort', 'HealthCheckProtocol',
                              'HTTPPort', 'Name', 'SecurityGroups', 'SSLCertificateId', 'Type'])
ALLOWED_HEALTH_CHECK_PROTOCOLS = frozenset(["HTTP", "HTTPS"])


def get_listeners(lb_name, target_group_name, subdomain, main_zone, configuration,
                  account_info: AccountArguments):
    ssl_cert = get_ssl_cert(subdomain, main_zone, configuration, account_info)
    return [{
        'Type': 'AWS::ElasticLoadBalancingV2::Listener',
        'Properties': {
            "Certificates": [{'CertificateArn': ssl_cert}],
            "Protocol": "HTTPS",
            "DefaultActions": [{'Type': 'forward', 'TargetGroupArn': {'Ref': target_group_name}}],
            'LoadBalancerArn': {'Ref': lb_name},
            "Port": 443
        }
    }]


def component_elastic_load_balancer_v2(definition,
                                       configuration: dict,
                                       args: TemplateArguments,
                                       info: dict,
                                       force,
                                       account_info: AccountArguments):
    lb_name = configuration["Name"]
    # domains pointing to the load balancer
    subdomain = ''
    main_zone = None
    for name, domain in configuration.get('Domains', {}).items():
        name = '{}{}'.format(lb_name, name)

        domain_name = "{0}.{1}".format(domain["Subdomain"], domain["Zone"])

        convert_cname_records_to_alias(domain_name)

        properties = {"Type": "A",
                      "Name": domain_name,
                      "HostedZoneName": domain["Zone"],
                      "AliasTarget": {"HostedZoneId": {"Fn::GetAtt": [lb_name,
                                                                      "CanonicalHostedZoneID"]},
                                      "DNSName": {"Fn::GetAtt": [lb_name, "DNSName"]}}}
        definition["Resources"][name] = {"Type": "AWS::Route53::RecordSet",
                                         "Properties": properties}

        if domain["Type"] == "weighted":
            definition["Resources"][name]["Properties"]['Weight'] = 0
            definition["Resources"][name]["Properties"]['SetIdentifier'] = "{0}-{1}".format(info["StackName"],
                                                                                            info["StackVersion"])
            subdomain = domain['Subdomain']
            main_zone = domain['Zone']  # type: str

    target_group_name = lb_name + 'TargetGroup'
    listeners = configuration.get('Listeners') or get_listeners(
        lb_name, target_group_name, subdomain, main_zone, configuration, account_info)

    health_check_protocol = configuration.get('HealthCheckProtocol') or 'HTTP'

    if health_check_protocol not in ALLOWED_HEALTH_CHECK_PROTOCOLS:
        raise click.UsageError('Protocol "{}" is not supported for LoadBalancer'.format(health_check_protocol))

    health_check_path = configuration.get("HealthCheckPath") or '/health'
    health_check_port = configuration.get("HealthCheckPort") or configuration["HTTPPort"]

    if configuration.get('NameSuffix'):
        version = '{}-{}'.format(info["StackVersion"],
                                 configuration['NameSuffix'])
        loadbalancer_name = get_load_balancer_name(info["StackName"], version)
        del(configuration['NameSuffix'])
    else:
        loadbalancer_name = get_load_balancer_name(info["StackName"],
                                                   info["StackVersion"])

    loadbalancer_scheme = configuration.get('Scheme') or 'internal'
    if loadbalancer_scheme == 'internet-facing':
        click.secho('You are deploying an internet-facing ELB that will be '
                    'publicly accessible! You should have OAUTH2 and HTTPS '
                    'in place!', bold=True, err=True)

    if loadbalancer_scheme not in ALLOWED_LOADBALANCER_SCHEMES:
        raise click.UsageError('Scheme "{}" is not supported for LoadBalancer'.format(loadbalancer_scheme))

    if loadbalancer_scheme == "internal":
        loadbalancer_subnet_map = "LoadBalancerInternalSubnets"
    else:
        loadbalancer_subnet_map = "LoadBalancerSubnets"

    tags = [
        # Tag "Name"
        {
            "Key": "Name",
            "Value": "{0}-{1}".format(info["StackName"], info["StackVersion"])
        },
        # Tag "StackName"
        {
            "Key": "StackName",
            "Value": info["StackName"],
        },
        # Tag "StackVersion"
        {
            "Key": "StackVersion",
            "Value": info["StackVersion"]
        }
    ]

    # load balancer
    definition["Resources"][lb_name] = {
        "Type": "AWS::ElasticLoadBalancingV2::LoadBalancer",
        "Properties": {
            'Name': loadbalancer_name,
            'Scheme': loadbalancer_scheme,
            'SecurityGroups': resolve_security_groups(configuration["SecurityGroups"], args.region),
            'Subnets': {"Fn::FindInMap": [loadbalancer_subnet_map, {"Ref": "AWS::Region"}, "Subnets"]},
            "Tags": tags
        }
    }
    definition["Resources"][target_group_name] = {
        'Type': 'AWS::ElasticLoadBalancingV2::TargetGroup',
        'Properties': {
            'Name': loadbalancer_name,
            'HealthCheckIntervalSeconds': '10',
            'HealthCheckPath': health_check_path,
            'HealthCheckPort': health_check_port,
            'HealthCheckProtocol': health_check_protocol,
            'HealthCheckTimeoutSeconds': '5',
            'HealthyThresholdCount': '2',
            'Port': configuration['HTTPPort'],
            'Protocol': 'HTTP',
            'UnhealthyThresholdCount': '2',
            'VpcId': account_info.VpcID,  # TODO: support multiple VPCs
            'Tags': tags,
            'TargetGroupAttributes': [{'Key': 'deregistration_delay.timeout_seconds', 'Value': '60'}]
        }
    }
    resource_names = set([lb_name, target_group_name])
    for i, listener in enumerate(listeners):
        if i == 0:
            suffix = ''
        else:
            suffix = str(i + 1)
        resource_name = lb_name + 'Listener' + suffix
        definition['Resources'][resource_name] = listener
        resource_names.add(resource_name)
    for key, val in configuration.items():
        # overwrite any specified properties, but only properties which were defined by us already
        for res in resource_names:
            if key in definition['Resources'][res]['Properties'] and key not in SENZA_PROPERTIES:
                definition['Resources'][res]['Properties'][key] = val
    return definition



def check_file_exceptions(stack_references: list):
    """
    Check all stack references to see if any references that looks like a yaml
    filename wasn't matched and raises an filerror
    """
    for stack in stack_references:
        stack.raise_file_exception()



from time import sleep

import boto3
from botocore.exceptions import ClientError

from .utils import extract_client_error_code

__all__ = ['BotoClientProxy']


class BotoClientProxy:
    def __init__(self, *args, **kwargs):
        self.__client = boto3.client(*args, **kwargs)

    @staticmethod
    def __decorator(function, *args, **kwargs):
        def wrapper(*args, **kwargs):
            max_tries = 5
            sleep_time = 10
            last_error = None
            for i in range(max_tries):
                try:
                    return function(*args, **kwargs)
                except ClientError as error:
                    last_error = error
                    if extract_client_error_code(error) == "Throttling":
                        sleep(sleep_time)
                    else:
                        raise
            else:
                if last_error:
                    raise last_error
        return wrapper

    def __getattr__(self, item):
        client_attr = getattr(self.__client, item)
        if callable(client_attr):
            return self.__decorator(client_attr)
        else:
            return client_attr




import codecs
import datetime

import yaml

from .manaus.boto_proxy import BotoClientProxy

LAUNCH_CONFIGURATION_PROPERTIES = set([
    'AssociatePublicIpAddress',
    'BlockDeviceMappings',
    'ClassicLinkVPCId',
    'ClassicLinkVPCSecurityGroups',
    'EbsOptimized',
    'IamInstanceProfile',
    'ImageId',
    'InstanceId',
    'InstanceMonitoring',
    'InstanceType',
    'KernelId',
    'KeyName',
    'LaunchConfigurationName',
    'PlacementTenancy',
    'RamdiskId',
    'SecurityGroups',
    'SpotPrice',
    'UserData',
])


def patch_user_data(old: str, new: dict):
    first_line, sep, data = old.partition('\n')
    data = yaml.safe_load(data)
    if not isinstance(data, dict):
        raise ValueError('Instance user data has invalid YAML: must be key/value pairs')
    data.update(**new)
    return first_line + sep + yaml.safe_dump(data, default_flow_style=False)


def patch_auto_scaling_group(group: dict, region: str, properties: dict):
    asg = BotoClientProxy('autoscaling', region)
    result = asg.describe_launch_configurations(LaunchConfigurationNames=[group['LaunchConfigurationName']])
    lcs = result['LaunchConfigurations']
    changed = False
    for lc in lcs:
        lc_props = {k: lc[k] for k in properties}
        if properties != lc_props:
            # create new launch configuration with specified properties
            kwargs = {}
            for key in LAUNCH_CONFIGURATION_PROPERTIES:
                # NOTE: we only take non-empty values (otherwise the parameter validation will complain :-( )
                val = lc.get(key)
                if val is not None and val != '':
                    if key == 'UserData':
                        val = codecs.decode(val.encode('utf-8'), 'base64').decode('utf-8')
                    kwargs[key] = val
            now = datetime.datetime.utcnow().strftime('%Y%m%dT%H%M%S')
            kwargs['LaunchConfigurationName'] = '{}-{}'.format(kwargs['LaunchConfigurationName'][:64], now)
            for key, val in properties.items():
                if key == 'UserData' and isinstance(val, dict):
                    kwargs[key] = patch_user_data(kwargs[key], val)
                else:
                    kwargs[key] = val
            asg.create_launch_configuration(**kwargs)
            asg.update_auto_scaling_group(AutoScalingGroupName=group['AutoScalingGroupName'],
                                          LaunchConfigurationName=kwargs['LaunchConfigurationName'])
            changed = True
    return changed



import sys
from collections import defaultdict
from copy import deepcopy
from enum import Enum
from typing import Any, Dict, Iterable, Iterator, List, Optional, Tuple, Union

from click import confirm

from .boto_proxy import BotoClientProxy
from .exceptions import (ELBNotFound, HostedZoneNotFound, InvalidState,
                         RecordNotFound)


class ChangeAction(str, Enum):
    """
    DNS record type
    """

    CREATE = 'CREATE'
    DELETE = 'DELETE'
    UPSERT = 'UPSERT'


class RecordType(str, Enum):
    """
    DNS record type
    """
    A = 'A'
    AAAA = 'AAAA'
    CNAME = 'CNAME'
    MX = 'MX'
    NS = 'NS'
    PRT = 'PRT'
    SOA = 'SOA'
    SPF = 'SPF'
    SRV = 'SRV'
    TXT = 'TXT'


class Route53HostedZone:
    """
    See:
    http://boto3.readthedocs.io/en/latest/reference/services/route53.html#Route53.Client.list_hosted_zones
    """

    def __init__(self,
                 id: str,
                 name: Optional[str]=None,
                 caller_reference: Optional[str]=None,
                 config: Optional[Dict[str, Any]]=None,
                 resource_record_set_count: Optional[int]=None):
        self.id = id
        self.name = name
        self.caller_reference = caller_reference
        self.config = config
        self.resource_record_set_count = resource_record_set_count

        # extra properties
        self.domain_name = name.rstrip('.') if name else None

    def __repr__(self):
        identifier = self.name or self.id
        return '<Route53HostedZone: {name}>'.format(name=identifier)

    @classmethod
    def get_by_domain_name(cls, domain_name: str) -> 'Route53HostedZone':
        hosted_zone_iter = Route53.get_hosted_zones(domain_name=domain_name)
        try:
            hosted_zone = next(hosted_zone_iter)
        except StopIteration:
            raise HostedZoneNotFound(domain_name)
        return hosted_zone

    @classmethod
    def get_by_id(cls, id: str) -> 'Route53HostedZone':
        hosted_zone_iter = Route53.get_hosted_zones(id=id)
        try:
            hosted_zone = next(hosted_zone_iter)
        except StopIteration:
            raise HostedZoneNotFound(id)
        return hosted_zone

    @classmethod
    def from_boto_dict(cls, hosted_zone_dict: Dict[str, Any]) -> 'Route53HostedZone':
        id = hosted_zone_dict['Id']
        name = hosted_zone_dict['Name']
        caller_reference = hosted_zone_dict['CallerReference']
        config = hosted_zone_dict['Config']
        resource_record_set_count = hosted_zone_dict['ResourceRecordSetCount']

        return cls(id, name, caller_reference, config,
                   resource_record_set_count)

    def change(self,
               changes: Iterable[Tuple[str, "Route53Record"]],
               comment: Optional[str] = None) -> Dict[str, Any]:
        """
        Submits changes to Route53. Returns the dict sent to
        AWS.

        See:
        http://boto3.readthedocs.io/en/latest/reference/services/route53.html#Route53.Client.change_resource_record_sets
        """

        client = BotoClientProxy('route53')

        change_batch = {'Changes': []}
        if comment is not None:
            change_batch['Comment'] = comment
        for action, record in changes:
            change = {'Action': action.upper(),
                      'ResourceRecordSet': record.boto_dict}
            change_batch['Changes'].append(change)

        if change_batch['Changes']:
            client.change_resource_record_sets(HostedZoneId=self.id,
                                               ChangeBatch=change_batch)

        return change_batch

    def create(self,
               changed_records: Iterable["Route53Record"],
               comment: Optional[str] = None) -> Dict[str, Any]:
        """
        Submits changes to be created in Route53. Returns the dict sent to
        AWS.

        See:
        http://boto3.readthedocs.io/en/latest/reference/services/route53.html#Route53.Client.change_resource_record_sets
        """

        changes = [(ChangeAction.CREATE, record) for record in changed_records]

        return self.change(changes, comment)

    def delete(self,
               changed_records: Iterable["Route53Record"],
               comment: Optional[str] = None) -> Dict[str, Any]:
        """
        Submits changes to be deleted in Route53. Returns the dict sent to
        AWS.

        See:
        http://boto3.readthedocs.io/en/latest/reference/services/route53.html#Route53.Client.change_resource_record_sets
        """

        changes = [(ChangeAction.DELETE, record) for record in changed_records]

        return self.change(changes, comment)

    def upsert(self,
               changed_records: Iterable["Route53Record"],
               comment: Optional[str]=None) -> Dict[str, Any]:
        """
        Submits changes to be upsterted in Route53. Returns the dict sent to
        AWS.

        See:
        http://boto3.readthedocs.io/en/latest/reference/services/route53.html#Route53.Client.change_resource_record_sets
        """

        changes = [(ChangeAction.UPSERT, record) for record in changed_records]

        return self.change(changes, comment)


class Route53Record:
    """
    See:
    http://boto3.readthedocs.io/en/latest/reference/services/route53.html#Route53.Client.list_resource_record_sets
    """

    def __init__(self,
                 name: str,
                 type: str,
                 ttl: Optional[int]=None,
                 resource_records: Optional[List[Dict[str, str]]]=None,
                 alias_target: Optional[Dict[str, Union[str, bool]]]=None,
                 failover: Optional[str]=None,
                 geo_location: Optional[Dict[str, str]]=None,
                 health_check_id: Optional[int]=None,
                 region: Optional[str]=None,
                 set_identifier: Optional[str]=None,
                 traffic_policy_instance_id: Optional[str]=None,
                 weight: Optional[int]=None,
                 hosted_zone: Optional["Route53HostedZone"]=None):

        self.name = name
        self.type = RecordType[type.upper()]

        self.ttl = ttl
        self.resource_records = resource_records
        self.alias_target = alias_target  # Alias resource record sets only
        self.failover = failover  # Failover resource record sets only
        self.geo_location = geo_location  # Geo location resource record sets only
        self.health_check_id = health_check_id  # Health Check resource record sets only
        self.region = region  # Latency-based resource record sets only
        self.set_identifier = set_identifier  # Weighted, Latency, Geo, and Failover resource record sets only
        self.traffic_policy_instance_id = traffic_policy_instance_id
        self.weight = weight  # Weighted resource record sets only

        self.hosted_zone = hosted_zone

    def __repr__(self):
        return '<Route53Record: {name}>'.format_map(vars(self))

    @classmethod
    def get_by_domain_name(cls, domain_name: str) -> "Route53Record":
        records_iter = Route53.get_records(name=domain_name)
        try:
            record = next(records_iter)
        except StopIteration:
            raise RecordNotFound(domain_name)
        return record

    @property
    def boto_dict(self):
        """
        Generates the dict to change records set

        See:
         http://boto3.readthedocs.io/en/latest/reference/services/route53.html#Route53.Client.change_resource_record_sets
        """

        boto_dict = {"Name": self.name,
                     "Type": self.type.value}

        optional_parameters = [('SetIdentifier', self.set_identifier),
                               ('Weight', self.weight),
                               ('Region', self.region),
                               ('GeoLocation', self.geo_location),
                               ('Failover', self.failover),
                               ('TTL', self.ttl),
                               ('ResourceRecords', self.resource_records),
                               ('AliasTarget', self.alias_target)]

        for key, value in optional_parameters:
            if value is not None:
                boto_dict[key] = value

        return boto_dict

    @classmethod
    def from_boto_dict(cls,
                       record_dict: Dict[str, Any],
                       hosted_zone: Optional["Route53HostedZone"] = None) -> 'Route53Record':
        """
        Returns a Route53Record based on the dict returned by boto3
        """
        name = record_dict['Name']
        type = record_dict['Type']

        ttl = record_dict.get('TTL')
        resource_records = record_dict.get('ResourceRecords')
        alias_target = record_dict.get('AliasTarget')
        failover = record_dict.get('Failover')
        geo_location = record_dict.get('GeoLocation')
        health_check_id = record_dict.get('HealthCheckId')
        region = record_dict.get('Region')
        set_identifier = record_dict.get('SetIdentifier')
        traffic_policy_instance_id = record_dict.get('TrafficPolicyInstanceId')
        weight = record_dict.get('Weight')

        return cls(name, type, ttl, resource_records,
                   alias_target, failover, geo_location, health_check_id,
                   region, set_identifier, traffic_policy_instance_id, weight,
                   hosted_zone=hosted_zone)

    def to_alias(self,
                 target_type: RecordType=RecordType.A) -> "Route53Record":
        """
        Converts record to an Alias record.

        By default it converts the record to an A type Alias but this can
        be overridden by provident a `target_type`.
        """

        # this needs to be imported locally to avoid circular imports
        from .elb import ELB

        if self.alias_target is not None:
            # Record is already an Alias
            return deepcopy(self)
        if self.type == RecordType.CNAME:
            dns_name = self.resource_records[0]['Value']
            # dns name looks like lb-name-123456.aws-region-1.elb.amazonaws.com
            sub_domain, _ = dns_name.split('.', maxsplit=1)  # type: str
            lb_name, _ = sub_domain.rsplit('-', maxsplit=1)

            if dns_name.endswith(".elb.amazonaws.com"):
                elb = ELB.get_by_dns_name(dns_name)
                hosted_zone_id = elb.hosted_zone.id
            else:
                _, hosted_zone_id = self.hosted_zone.id.rsplit('/', maxsplit=1)

            alias_target = {"HostedZoneId": hosted_zone_id,
                            "DNSName": dns_name,
                            "EvaluateTargetHealth": False}
            return self.__class__(name=self.name,
                                  type=target_type,
                                  alias_target=alias_target,
                                  traffic_policy_instance_id=self.traffic_policy_instance_id,
                                  set_identifier=self.set_identifier,
                                  weight=self.weight)
        else:
            raise NotImplementedError("Conversion of {} records to Alias is "
                                      "still not implemented".format(self.type))


class Route53:

    def __init__(self):
        self.client = BotoClientProxy('route53')

    @staticmethod
    def get_hosted_zones(domain_name: Optional[str]=None,
                         id: Optional[str]=None) -> Iterator[Route53HostedZone]:
        """
        Gets hosted zones from Route53. If a ``domain_name`` is provided
        only hosted zones that match the domain name will be yielded
        """

        if domain_name is not None:
            domain_name = '{}.'.format(domain_name.rstrip('.'))

        client = BotoClientProxy('route53')
        result = client.list_hosted_zones()
        hosted_zones = result["HostedZones"]
        while result.get('IsTruncated', False):
            recordfilter = {'Marker': result['NextMarker']}
            result = client.list_hosted_zones(**recordfilter)
            hosted_zones.extend(result['HostedZones'])

        for zone in hosted_zones:
            hosted_zone = Route53HostedZone.from_boto_dict(zone)

            if domain_name and not domain_name.endswith(hosted_zone.name):
                continue

            if id and hosted_zone.id != id:
                continue

            yield hosted_zone

    @classmethod
    def get_records(cls, *,
                    name: Optional[str]=None) -> Iterator[Route53Record]:
        client = BotoClientProxy('route53')
        if name is not None and not name.endswith('.'):
            name += '.'
        for zone in cls.get_hosted_zones():
            response = client.list_resource_record_sets(HostedZoneId=zone.id)
            resources = response["ResourceRecordSets"]  # type: List[Dict[str, Any]]

            # If the response includes more than maxitems resource record sets,
            # the value of the IsTruncated element in the response is true,
            # and the values of the NextRecordName and NextRecordType elements
            # in the response identify the first resource record set in the
            # next group of maxitems resource record sets.
            while response.get('IsTruncated', False):
                next_name = response['NextRecordName']
                next_type = response['NextRecordType']
                response = client.list_resource_record_sets(HostedZoneId=zone.id,
                                                            StartRecordName=next_name,
                                                            StartRecordType=next_type)
                resources.extend(response['ResourceRecordSets'])

            for resource in resources:
                record = Route53Record.from_boto_dict(resource,
                                                      hosted_zone=zone)
                if name is not None and record.name != name:
                    continue
                yield record


def convert_cname_records_to_alias(domain_name: str):
    records = Route53.get_records(name=domain_name)
    converted_records = defaultdict(lambda: {'delete': [],
                                             'upsert': []})
    for record in records:
        if record.type == RecordType.CNAME:
            converted_records[record.hosted_zone]['delete'].append(record)
            try:
                alias_record = record.to_alias()
                converted_records[record.hosted_zone]['upsert'].append(alias_record)
            except ELBNotFound:
                print("Dangling DNS record  will be deleted: {}".format(record.name))

    if converted_records:
        for hosted_zone, records in converted_records.items():
            s = 's' if records != 1 else ''
            interactive_terminal = sys.stdin.isatty()

            msg = ("\n  {name} ({hz}): {n} record{s} need to be converted to "
                   "Alias records".format(name=domain_name,
                                          hz=hosted_zone.name,
                                          n=len(records['upsert']),
                                          s=s))

            # if we are not in an interactive terminal change without asking
            # this is not ideal, but it's needed because of lizzy
            if not interactive_terminal or confirm(msg):
                hosted_zone.delete(records['delete'],
                                   comment="Records that will be converted "
                                           "to Alias")

                print("  Deleted old record{s}".format(s=s))

                hosted_zone.upsert(records['upsert'],
                                   comment="Convert non alias records")
                print("  Inserted alias record{s}".format(s=s))
            else:
                raise InvalidState("Can't change domains because there are "
                                   "non A Type records.")



"""
Functions to scale and respawn Auto Scale Groups
"""
import time

from clickclick import Action, info

from .manaus.boto_proxy import BotoClientProxy

SCALING_PROCESSES_TO_SUSPEND = ['AZRebalance', 'AlarmNotification', 'ScheduledActions']
RUNNING_LIFECYCLE_STATES = set(['Pending', 'InService', 'Rebooting'])


def get_auto_scaling_group(asg, asg_name: str):
    '''Get boto3 Auto Scaling Group by name or raise exception'''
    result = asg.describe_auto_scaling_groups(AutoScalingGroupNames=[asg_name])
    groups = result['AutoScalingGroups']
    if not groups:
        raise Exception('Auto Scaling Group {} not found'.format(asg_name))
    return groups[0]


def get_instances_to_terminate(group, desired_launch_config: str, force: bool):
    '''Return set of instance IDs to terminate for given Auto Scaling Group

    Returns all instances where the launch configuration is not up-to-date'''
    instances_to_terminate = set()
    instances_ok = set()
    for instance in group['Instances']:
        if instance['LifecycleState'] in RUNNING_LIFECYCLE_STATES:
            # NOTE: LaunchConfigurationName key might be missing (if config was deleted..)
            if not force and instance.get('LaunchConfigurationName') == desired_launch_config:
                instances_ok.add(instance['InstanceId'])
            else:
                instances_to_terminate.add(instance['InstanceId'])
    return instances_to_terminate, instances_ok


def get_instances_in_service(group, region: str):
    '''Get set of instance IDs with ELB "InService" state'''
    instances_in_service = set()
    # TODO: handle auto scaling groups without any ELB
    lb_names = group['LoadBalancerNames']
    if lb_names:
        # check ELB status
        elb = BotoClientProxy('elb', region)
        for lb_name in lb_names:
            result = elb.describe_instance_health(LoadBalancerName=lb_name)
            for instance in result['InstanceStates']:
                if instance['State'] == 'InService':
                    instances_in_service.add(instance['InstanceId'])
    else:
        # just use ASG LifecycleState
        group = get_auto_scaling_group(BotoClientProxy('autoscaling', region),
                                       group['AutoScalingGroupName'])
        for instance in group['Instances']:
            if instance['LifecycleState'] == 'InService':
                instances_in_service.add(instance['InstanceId'])
    return instances_in_service


def scale_out(asg, asg_name, region: str, new_min_size, new_max_size, new_desired_capacity: int):
    """
    Scale out the given Auto Scaling Group to the given capacity
    and wait for all instances to become "InService" in ELB
    """
    with Action('Scaling to {} instances..'.format(new_desired_capacity)) as act:
        asg.update_auto_scaling_group(AutoScalingGroupName=asg_name,
                                      MinSize=new_min_size,
                                      MaxSize=new_max_size,
                                      DesiredCapacity=new_desired_capacity)
        while True:
            current_group = get_auto_scaling_group(asg, asg_name)
            instances_in_service = get_instances_in_service(current_group, region)
            if len(instances_in_service) >= new_desired_capacity:
                break
            time.sleep(5)
            act.progress()
    return current_group


def terminate_instance(asg, region: str, group, instance: str):
    """
    Terminate a single EC2 instance in given Auto Scaling Group and wait for
    it to become "OutOfService"
    """

    with Action('Terminating old instance {}..'.format(instance)) as act:
        asg.terminate_instance_in_auto_scaling_group(InstanceId=instance,
                                                     ShouldDecrementDesiredCapacity=False)
        instances_in_service = get_instances_in_service(group, region)
        while instance in instances_in_service:
            time.sleep(2)
            act.progress()
            instances_in_service = get_instances_in_service(group, region)


def do_respawn_auto_scaling_group(asg_name: str, group: dict, region: str,
                                  instances_to_terminate: set, inplace: bool):
    """
    Respawn ASG.
    """
    asg = BotoClientProxy('autoscaling', region)
    with Action('Suspending scaling processes for {}..'.format(asg_name)):
        asg.suspend_processes(AutoScalingGroupName=asg_name,
                              ScalingProcesses=SCALING_PROCESSES_TO_SUSPEND)
    extra_capacity = 0 if inplace else 1
    new_min_size = group['MinSize'] + extra_capacity
    new_max_size = group['MaxSize'] + extra_capacity
    new_desired_capacity = group['DesiredCapacity'] + extra_capacity
    # TODO: error handling (rollback in case of exception?)
    while instances_to_terminate:
        current_group = scale_out(asg, asg_name, region, new_min_size,
                                  new_max_size, new_desired_capacity)
        instance = sorted(instances_to_terminate)[0]
        terminate_instance(asg, region, current_group, instance)
        instances_to_terminate.remove(instance)

    with Action('Resetting Auto Scaling Group to original capacity '
                '({MinSize}-{DesiredCapacity}-{MaxSize})..'.format_map(group)):
        asg.update_auto_scaling_group(AutoScalingGroupName=asg_name,
                                      MinSize=group['MinSize'],
                                      MaxSize=group['MaxSize'],
                                      DesiredCapacity=group['DesiredCapacity'])

    with Action('Resuming scaling processes for {}..'.format(asg_name)):
        asg.resume_processes(AutoScalingGroupName=asg_name)


def respawn_auto_scaling_group(asg_name: str, region: str, inplace: bool=False, force: bool=False):
    """
    Respawn all EC2 instances in the Auto Scaling Group whose launch
    configuration is not up-to-date
    """
    asg = BotoClientProxy('autoscaling', region)
    group = get_auto_scaling_group(asg, asg_name)
    desired_launch_config = group['LaunchConfigurationName']
    instances_to_terminate, instances_ok = get_instances_to_terminate(group,
                                                                      desired_launch_config,
                                                                      force)
    info('{}/{} instances need to be updated in {}'.format(
        len(instances_to_terminate),
        len(instances_to_terminate) + len(instances_ok),
        asg_name
    ))
    if instances_to_terminate:
        do_respawn_auto_scaling_group(asg_name,
                                      group,
                                      region,
                                      instances_to_terminate,
                                      inplace)
    else:
        info('Nothing to do')



import boto3.session
import click

from .error_handling import HandleExceptions


def validate_region(ctx, param, value):
    """Validate Click region param parameter."""

    if value is not None:
        session = boto3.session.Session()
        valid_regions = session.get_available_regions('cloudformation')
        if value not in valid_regions:
            valid_regions.sort()
            raise click.BadParameter("'{}'. Region must be one of the "
                                     "following AWS regions:\n"
                                     "  - {}".format(value,
                                                     "\n  - ".join(valid_regions)))
    return value


def set_stacktrace_visible(ctx, param, value):
    HandleExceptions.stacktrace_visible = value


region_option = click.option('--region',
                             envvar='AWS_DEFAULT_REGION',
                             metavar='AWS_REGION_ID',
                             help='AWS region ID (e.g. eu-west-1)',
                             callback=validate_region)

parameter_file_option = click.option('--parameter-file',
                                     help='Config file for params',
                                     metavar='PATH')

output_option = click.option('-o', '--output',
                             type=click.Choice(['text', 'json', 'tsv']),
                             default='text',
                             help='Use alternative output format')

json_output_option = click.option('-o', '--output',
                                  type=click.Choice(['json', 'yaml']),
                                  default='json',
                                  help='Use alternative output format')

stacktrace_visible_option = click.option('--stacktrace-visible',
                                         is_flag=True,
                                         callback=set_stacktrace_visible,
                                         expose_value=False,
                                         help='Show stack trace instead of '
                                              'storing it')

watch_option = click.option('-W',
                            is_flag=True,
                            help='Auto update the screen every 2 seconds')

watchrefresh_option = click.option('-w', '--watch',
                                   type=click.IntRange(1, 300),
                                   metavar='SECS',
                                   help='Auto update the screen every X seconds')
GLOBAL_OPTIONS = {}



from types import ModuleType, FunctionType
import pkg_resources


def get_template_description(name, module: ModuleType):
    return '{}: {}'.format(name, (module.__doc__ or "").strip())


def has_functions(module, names):
    return all(isinstance(getattr(module, function_name, None), FunctionType) for function_name in names)


def get_templates() -> dict:
    """
    Returns a dict with all the template modules
    """
    entry_points = pkg_resources.iter_entry_points('senza.templates')
    template_modules = {}
    for e in entry_points:  # type: pkg_resources.EntryPoint
        try:
            module = e.resolve()
        except ImportError:
            # ignore bad entry points
            continue
        else:
            # make sure the entry point resolves to a module with the essential interface functions
            if (isinstance(module, ModuleType) and
               has_functions(module, ('gather_user_variables', 'generate_definition'))):
                template_modules[e.name] = module
    return template_modules



"""
Functions to handle exceptions that bubble to the top, including Sentry
integration
"""

import sys
from tempfile import NamedTemporaryFile
from traceback import format_exception
from typing import Any, Dict, Optional  # noqa: F401

import senza
import yaml.constructor
from botocore.exceptions import ClientError, NoCredentialsError
from clickclick import fatal_error
from raven import Client

from .configuration import configuration
from .exceptions import InvalidDefinition, PiuNotFound
from .manaus.exceptions import (ELBNotFound, HostedZoneNotFound, InvalidState,
                                RecordNotFound)
from .manaus.utils import extract_client_error_code


def store_exception(exception: Exception) -> str:
    """
    Stores the exception in a temporary file and returns its filename
    """

    tracebacks = format_exception(etype=type(exception),
                                  value=exception,
                                  tb=exception.__traceback__)  # type: [str]

    content = ''.join(tracebacks)

    with NamedTemporaryFile(prefix="senza-traceback-", delete=False) as error_file:
        file_name = error_file.name
        error_file.write(content.encode())

    return file_name


def is_credentials_expired_error(client_error: ClientError) -> bool:
    """Return true if the exception's error code is ExpiredToken or RequestExpired"""
    return extract_client_error_code(client_error) in ['ExpiredToken',
                                                       'RequestExpired']


def is_access_denied_error(client_error: ClientError) -> bool:
    """
    Checks the ``ClientError`` details to find out if it is an
    Access Denied Error
    """
    return extract_client_error_code(client_error) in ['AccessDenied']


def is_validation_error(client_error: ClientError) -> bool:
    """
    Checks the ``ClientError`` details to find out if it is an
    Validation Error
    """
    return extract_client_error_code(client_error) == 'ValidationError'


def die_fatal_error(message):
    """Sent error message to stderr, in red, and exit"""
    fatal_error(message, err=True)


class HandleExceptions:
    """Class HandleExceptions will display various error messages
    depending on the type of the exception and show the stacktrack for general exceptions
    depending on the value of stacktrace_visible"""

    stacktrace_visible = False

    def __init__(self, function):
        self.function = function

    def die_unknown_error(self, unknown_exception: Exception):
        """
        Handles unknown exceptions, shipping them to sentry if it's configured.

        If stacktrace_visible the stacktrace will be printed otherwise the
        stacktrace will be stored in a temporary file or sent to sentry.
        """
        if sentry:
            # The exception should always be sent to sentry if sentry is
            # configured
            sentry.captureException()
        if self.stacktrace_visible:
            raise unknown_exception
        elif sentry:
            die_fatal_error("Unknown Error: {e}.\n"
                            "This error will be pushed to sentry ".format(e=unknown_exception))
        elif not sentry:
            file_name = store_exception(unknown_exception)
            die_fatal_error('Unknown Error: {e}.\n'
                            'Please create an issue with the '
                            'content of {fn}'.format(e=unknown_exception,
                                                     fn=file_name))

    def __call__(self, *args, **kwargs):
        try:
            self.function(*args, **kwargs)
        except NoCredentialsError:
            die_fatal_error(
                'No AWS credentials found. Use the "mai" command-line tool '
                'to get a temporary access key\n'
                'or manually configure either ~/.aws/credentials '
                'or AWS_ACCESS_KEY_ID/AWS_SECRET_ACCESS_KEY.')
        except ClientError as client_error:
            sys.stdout.flush()
            if is_credentials_expired_error(client_error):
                die_fatal_error('AWS credentials have expired.\n'
                                'Use the "mai" command line tool to get a new '
                                'temporary access key.')
            elif is_access_denied_error(client_error):
                die_fatal_error(
                    "AWS missing access rights.\n{}".format(
                        client_error.response['Error']['Message']))
            elif is_validation_error(client_error):
                die_fatal_error(
                    "Validation Error: {}".format(
                        client_error.response['Error']['Message']))
            else:
                self.die_unknown_error(client_error)
        except yaml.constructor.ConstructorError as yaml_error:
            err_mesg = "Error parsing definition file:\n{}".format(yaml_error)
            if yaml_error.problem == "found unhashable key":
                err_mesg += "Please quote all variable values"
            die_fatal_error(err_mesg)
        except PiuNotFound as error:
            die_fatal_error(
                "{}\nYou can install piu with the following command:"
                "\nsudo pip3 install --upgrade stups-piu".format(error))
        except (ELBNotFound, HostedZoneNotFound, RecordNotFound,
                InvalidDefinition, InvalidState) as error:
            die_fatal_error(error)
        except Exception as unknown_exception:
            # Catch All
            self.die_unknown_error(unknown_exception)


def setup_sentry(sentry_endpoint: Optional[str]):
    """
    This function setups sentry, this exists mostly to make sentry integration
    easier to test
    """
    if sentry_endpoint is not None:
        sentry_client = Client(sentry_endpoint,
                               release=senza.__version__)
    else:
        sentry_client = None

    return sentry_client

sentry = setup_sentry(configuration.get('sentry.endpoint'))



from typing import Optional

from click import argument, command
from click.exceptions import BadArgumentUsage

from ..configuration import configuration
from ..exceptions import InvalidConfigKey


@command('config')
@argument('key')
@argument('value', required=False)
def cmd_config(key: str, value: Optional[str]):
    """
    Get and set senza options.
    """
    if value is None:
        try:
            value = configuration[key]
            print(value)
        except InvalidConfigKey as e:
            raise BadArgumentUsage(e)
        except KeyError:
            exit(1)
    else:
        try:
            configuration[key] = value
        except InvalidConfigKey as e:
            raise BadArgumentUsage(e)
